from numpy.random import RandomState
import numpy
from nbodykit.utils import FrontPadArray

class MPIRandomState:
    """ A Random number generator that is invariant against number of ranks,
        when the total size of random number requested is kept the same.

        The algorithm here assumes the random number generator from numpy
        produces uncorrelated results when the seeds are sampled from a single
        RNG.

        The sampler methods are collective calls; multiple calls will return
        uncorrerlated results.

        The result is only invariant under diif comm.size when allreduce(size)
        and chunksize are kept invariant.

    """
    def __init__(self, comm, seed, size, chunksize=100000):
        self.comm = comm
        self.seed = seed
        self.chunksize = chunksize

        self.size = size
        self.csize = numpy.sum(comm.allgather(size), dtype='intp')

        self._start = numpy.sum(comm.allgather(size)[:comm.rank], dtype='intp')
        self._end = self._start + self.size

        self._first_ichunk = self._start // chunksize

        self._skip = self._start - self._first_ichunk * chunksize

        nchunks = (comm.allreduce(numpy.array(size, dtype='intp')) + chunksize - 1) // chunksize
        self.nchunks = nchunks

        self._serial_rng = RandomState(seed)

    def _prepare_args_and_result(self, args, itemshape, dtype):
        """ pad every item in args with values from previous ranks,
            and create an array for holding the result with the same length.

            Returns
            -------
            padded_r, padded_args

        """

        r = numpy.zeros((self.size,) + tuple(itemshape), dtype=dtype)

        r_and_args = (r,) + tuple(args)
        r_and_args_b = numpy.broadcast_arrays(*r_and_args)

        padded = []

        # we don't need to pad scalars,
        # loop over broadcasted and non broadcast version to figure this out)
        for a, a_b in zip(r_and_args, r_and_args_b):
            if numpy.isscalar(a):
                # use the scalar, no need to pad.
                padded.append(a)
            else:
                # not a scalar, pad
                padded.append(FrontPadArray(a_b, self._skip, self.comm))

        return padded[0], padded[1:]

    def poisson(self, lam, itemshape=(), dtype='f8'):
        """ Produce `self.size` poissons, each of shape itemshape. This is a collective MPI call. """
        def sampler(rng, args, size):
            lam, = args
            return rng.poisson(lam=lam, size=size)
        return self._call_rngmethod(sampler, (lam,), itemshape, dtype)

    def choice(self, choices, itemshape=(), replace=True, p=None):
        """ Produce `self.size` choices, each of shape itemshape. This is a collective MPI call. """
        dtype = numpy.array(choices).dtype
        def sampler(rng, args, size):
            return rng.choice(choices, size=size, replace=replace, p=p)

        return self._call_rngmethod(sampler, (), itemshape, dtype)

    def normal(self, loc=0, scale=1, itemshape=(), dtype='f8'):
        """ Produce `self.size` normals, each of shape itemshape. This is a collective MPI call. """
        def sampler(rng, args, size):
            loc, scale = args
            return rng.normal(loc=loc, scale=scale, size=size)
        return self._call_rngmethod(sampler, (loc, scale), itemshape, dtype)

    def uniform(self, low=0., high=1.0, itemshape=(), dtype='f8'):
        """ Produce `self.size` uniforms, each of shape itemshape. This is a collective MPI call. """
        def sampler(rng, args, size):
            low, high = args
            return rng.uniform(low=low, high=high,size=size)
        return self._call_rngmethod(sampler, (low, high), itemshape, dtype)

    def _call_rngmethod(self, sampler, args, itemshape, dtype='f8'):
        """
            Loop over the seed table, and call sampler(rng, args, size)
            on each rng, with matched input args and size.

            the args are padded in the front such that the rng is invariant
            no matter how self.size is distributed.

            truncate the return value at the front to match the requested `self.size`.
        """

        seeds = self._serial_rng.randint(0, high=0xffffffff, size=self.nchunks)

        padded_r, running_args = self._prepare_args_and_result(args, itemshape, dtype)

        running_r = padded_r
        ichunk = self._first_ichunk

        while len(running_r) > 0:
            # at most get a full chunk, or the remaining items
            nreq = min(len(running_r), self.chunksize)

            seed = seeds[ichunk]
            rng = RandomState(seed)
            args = tuple([a if numpy.isscalar(a) else a[:nreq] for a in running_args])

            # generate nreq random items from the sampler 
            chunk = sampler(rng, args=args,
                size=(nreq,) + tuple(itemshape))

            running_r[:nreq] = chunk

            # update running arrays, since we have finished nreq items
            running_r = running_r[nreq:]
            running_args = tuple([a if numpy.isscalar(a) else a[nreq:] for a in running_args])

            ichunk = ichunk + 1

        return padded_r[self._skip:]




from nbodykit.base.catalog import CatalogSourceBase
from nbodykit.utils import attrs_to_dict

import numpy
import logging
from six import string_types


class MultipleSpeciesCatalog(CatalogSourceBase):
    """
    A CatalogSource interface for handling multiples species
    of particles.

    This CatalogSource stores a copy of the original CatalogSource objects
    for each species, providing access to the columns via the format
    ``species/`` where "species" is one of the species names provided.

    Parameters
    ----------
    names : list of str
        list of strings specifying the names of the various species;
        data columns are prefixed with "species/" where "species" is
        in ``names``
    *species : two or more CatalogSource objects
        catalogs to be combined into a single catalog, which give the
        data for different species of particles; as many catalogs
        as names must be provided

    Examples
    --------

    Initialization:

    >>> data = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42)
    >>> randoms = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84)
    >>> cat = MultipleSpeciesCatalog(['data', 'randoms'], data, randoms)

    Accessing the Catalogs for individual species:

    >>> data = cat["data"] # a copy of the original "data" object

    Accessing individual columns:

    >>> data_pos = cat["data/Position"]

    Setting new columns:

    >>> cat["data"]["new_column"] = 1.0
    >>> assert "data/new_column" in cat

    """
    logger = logging.getLogger('MultipleSpeciesCatalog')

    def __repr__(self):
        return "MultipleSpeciesCatalog(species=%s)" %str(self.attrs['species'])

    def __init__(self, names, *species, **kwargs):

        if len(set(names)) != len(names):
            raise ValueError("each species must have a unique name")
        if not all(cat.comm is species[0].comm for cat in species):
            raise ValueError("communicator mismatch in MultipleSpeciesCatalog")
        if len(names) != len(species):
            raise ValueError("a name must be provided for each species catalog provided")

        CatalogSourceBase.__init__(self, species[0].comm)

        self.attrs['species'] = names

        # update the dictionary with data/randoms attrs
        for cat, name in zip(species, names):
            self.attrs.update(attrs_to_dict(cat, name + '.'))

        # update the rest of meta-data
        self.attrs.update(kwargs)

        # no size!
        self.size = NotImplemented
        self.csize = NotImplemented

        # store copies of the original input catalogs as (name:catalog) dict
        self._sources = {name:cat.copy() for name,cat in zip(names, species)}

    @property
    def species(self):
        """
        List of species names
        """
        return self.attrs['species']

    @property
    def columns(self):
        """
        Columns for individual species can be accessed using a ``species/``
        prefix and the column name, i.e., ``data/Position``.
        """
        return ['%s/%s' %(species, col)
                for species in self.species
                for col in self._sources[species].columns]

    @property
    def hardcolumns(self):
        """
        Hardcolumn of the form ``species/name``
        """
        return ['%s/%s' %(species, col)
                for species in self.species
                for col in self._sources[species].hardcolumns]

    def __getitem__(self, key):
        """
        This provides access to the underlying data in two ways:

        - The CatalogSource object for a species can be accessed if ``key``
          is a species name.
        - Individual columns for a species can be accessed using the
          format: ``species/column``.
        """
        # return a new CatalogSource holding only the specific species
        if isinstance(key, string_types):
            if key in self.species:
                return self._sources[key]

            species, subcol = split_column(key, self.species)
            return CatalogSourceBase.__getitem__(self._sources[species], subcol)

        return CatalogSourceBase.__getitem__(self, key)

    def __setitem__(self, col, value):
        """
        Add columns to any of the species catalogs.

        .. note::
            New column names should be prefixed by 'species/' where
            'species' is a name in the :attr:`species` attribute.
        """
        species, subcol = split_column(col, self.species)

        # check size
        size = len(self._sources[species])
        if not numpy.isscalar(value):
            if len(value) != size:
                args = (col, size, len(value))
                raise ValueError("error setting '%s' column, data must be array of size %d, not %d" % args)

        # add the column to the CatalogSource in "_sources"
        return CatalogSourceBase.__setitem__(self._sources[species], subcol, value)

    def __delitem__(self, col):
        """
        Delete a column of the form ``species/column``
        """
        species, subcol = split_column(col, self.species)
        return CatalogSourceBase.__delitem__(self._sources[species], subcol)


    def to_mesh(self, Nmesh=None, BoxSize=None, dtype='f4', interlaced=False,
                compensated=False, resampler='cic', weight='Weight',
                value='Value', selection='Selection', position='Position', window=None):
        """
        Convert the catalog to a mesh, which knows how to "paint" the
        the combined density field, summed over all particle species.

        Parameters
        ----------
        Nmesh : int, 3-vector, optional
            the number of cells per box side; can be inferred from ``attrs``
            if the value is the same for all species
        BoxSize : float, 3-vector, optional
            the size of the box; can be inferred from ``attrs``
            if the value is the same for all species
        dtype : str, dtype, optional
            the data type of the mesh when painting
        interlaced : bool, optional
            whether to use interlacing to reduce aliasing when painting the
            particles on the mesh
        compensated : bool, optional
            whether to apply a Fourier-space transfer function to account for
            the effects of the gridding + aliasing
        resampler: str, optional
            the string name of the resampler to use when interpolating the
        weight : str, optional
            the name of the column specifying the weight for each particle
        selection : str, optional
            the name of the column that specifies which (if any) slice
            of the CatalogSource to take
        value: str, optional
            the name of the column specifying the field value for each particle
        position : str, optional
            the name of the column that specifies the position data of the
            objects in the catalog
        window : str, optional
            the string name of the window to use when interpolating  (deprecated, use resampler)
        """
        from nbodykit.source.mesh.species import MultipleSpeciesCatalogMesh

        if window is not None:
            raise RuntimeError("use resampler instead")

        # verify that all of the required columns exist
        for name in self.species:
            for col in [position, selection, weight, value]:
                if col not in self[name]:
                    raise ValueError("the '%s' species is missing the '%s' column" %(name, col))

        # try to find BoxSize and Nmesh
        if BoxSize is None:
            BoxSize = check_species_metadata('BoxSize', self.attrs, self.species)
        if Nmesh is None:
            Nmesh = check_species_metadata('Nmesh', self.attrs, self.species)

        # return the mesh
        return MultipleSpeciesCatalogMesh(self,
                                          Nmesh=Nmesh,
                                          BoxSize=BoxSize,
                                          dtype=dtype,
                                          selection=selection,
                                          position=position,
                                          weight=weight,
                                          value=value,
                                          interlaced=interlaced,
                                          compensated=compensated,
                                          resampler=resampler)


def check_species_metadata(name, attrs, species):
    """
    Check to see if there is a single value for
    ``name`` in the meta-data of all the species
    """
    _vals = []

    if name in attrs:
        _vals.append(attrs[name])

    for s in species:
        if s + '.' + name in attrs:
            _vals.append(attrs.get(s + '.' + name ))

    if len(_vals) == 0:
        raise ValueError("please specify ``%s`` attributes" % name)

    if not all(numpy.equal(_vals[0], i).all() for i in _vals):
        raise ValueError("please specify ``%s`` attributes that are consistent for each species and for the multi species catalog; " % name)

    return _vals[0]

def split_column(col, species):
    """
    Split the column name of the form 'species/name'
    """
    fields = col.split('/')
    if len(fields) != 2:
        msg = "new column names should be prefixed by 'species/' where "
        msg += "'species' is one of %s" % str(species)
        raise ValueError(msg)

    this_species, subcol = fields
    if this_species not in species:
        args = (this_species, str(species))
        raise ValueError("species '%s' is not valid; should be one of %s" %args)
    return this_species, subcol



from nbodykit.source.catalog.species import MultipleSpeciesCatalog
from nbodykit.transform import ConstantArray

import numpy
import logging

def FKPWeightFromNbar(P0, nbar):
    """ Create FKPWeight from nbar, the number density of objects per redshift.

        Parameters
        ----------
        P0 : float
            the FKP normalization, when P0 == 0, returns 1.0, ignoring size / shape of nbar.

        nbar : array_like
            the number density of objects per redshift

        Returns
        -------
        FKPWeight : the FKPWeight, can be assigned to a catalog as a column
        to be consumed :class:`ConvolvedFFTPower`

    """
    if P0 != 0:
        return 1.0 / (1. + P0 * nbar)

    return 1.0


class FKPCatalog(MultipleSpeciesCatalog):
    """
    An interface for simultaneous modeling of a ``data`` CatalogSource and a
    ``randoms`` CatalogSource, in the spirit of
    `Feldman, Kaiser, and Peacock, 1994 <https://arxiv.org/abs/astro-ph/9304022>`_.

    This main functionality of this class is:

    *   provide a uniform interface to accessing columns from the
        ``data`` CatalogSource and ``randoms`` CatalogSource, using
        column names prefixed with "data/" or "randoms/"
    *   compute the shared :attr:`BoxSize` of the source, by
        finding the maximum Cartesian extent of the ``randoms``
    *   provide an interface to a mesh object, which knows how to paint the
        FKP density field from the ``data`` and ``randoms``

    Parameters
    ----------
    data : CatalogSource
        the CatalogSource of particles representing the `data` catalog
    randoms : CatalogSource, or None
        the CatalogSource of particles representing the `randoms` catalog
        if None is given an empty catalog is used.
    BoxSize : float, 3-vector, optional
        the size of the Cartesian box to use for the unified `data` and
        `randoms`; if not provided, the maximum Cartesian extent of the
        `randoms` defines the box
    BoxPad : float, 3-vector, optional
        optionally apply this additional buffer to the extent of the
        Cartesian box
    nbar : str, optional
        the name of the column specifying the number density as a function
        of redshift. default is NZ.
    P0 : float or None
        if not None, a column named FKPWeight is added to data and random based on nbar.

    References
    ----------
    - `Feldman, Kaiser, and Peacock, 1994 <https://arxiv.org/abs/astro-ph/9304022>`__
    """
    logger = logging.getLogger('FKPCatalog')

    def __repr__(self):
        return "FKPCatalog(species=%s)" %str(self.attrs['species'])

    def __init__(self, data, randoms, BoxSize=None, BoxPad=0.02, P0=None, nbar='NZ'):

        if randoms is None:
            # create an empty catalog.
            randoms = data[:0]

        # init the base class
        MultipleSpeciesCatalog.__init__(self, ['data', 'randoms'], data, randoms)

        for i, name in enumerate(self.species):
            if nbar not in self[name]:
                raise ValueError("Column `%s` is not defined in `%s`" % (nbar, name))

        self.nbar = nbar

        if P0 is not None:
            # create a default FKP weight column, based on nbar
            for i, name in enumerate(self.species):
                self[name]['FKPWeight'] = FKPWeightFromNbar(P0, self[name][self.nbar])
        else:
            # add a default FKP weight columns, based on nbar
            for i, name in enumerate(self.species):
                if 'FKPWeight' not in self[name]:
                    self[name]['FKPWeight'] = 1.0

        # determine the BoxSize
        if numpy.isscalar(BoxSize):
            BoxSize = numpy.ones(3)*BoxSize
        self.attrs['BoxSize'] = BoxSize
        if numpy.isscalar(BoxPad):
            BoxPad = numpy.ones(3)*BoxPad
        self.attrs['BoxPad'] = BoxPad

    def _define_bbox(self, position, selection, species):
        """
        Internal function to put the :attr:`randoms` CatalogSource in a
        Cartesian bounding box, using the positions of the given species.

        This function computings the size and center of the bounding box.

        #. `BoxSize` : array_like, (3,)
            if not provided, the BoxSize in each direction is computed from
            the maximum extent of the Cartesian coordinates of the :attr:`randoms`
            Source, with an optional, additional padding
        #. `BoxCenter`: array_like, (3,)
            the mean coordinate value in each direction; this is used to re-center
            the Cartesian coordinates of the :attr:`data` and :attr:`randoms`
            to the range of ``[-BoxSize/2, BoxSize/2]``

        """
        from nbodykit.utils import get_data_bounds

        # compute the min/max of the position data
        pos, sel = self[species].read([position, selection])
        pos_min, pos_max = get_data_bounds(pos, self.comm, selection=sel)

        if self.comm.rank == 0:
            self.logger.info("cartesian coordinate range: %s : %s" %(str(pos_min), str(pos_max)))

        if numpy.isinf(pos_min).any() or numpy.isinf(pos_max).any():
            raise ValueError("Range of positions from `%s` is infinite;"
                    "try to use the other species with (bbox_from_species='data'." % species)

        # used to center the data in the first cartesian quadrant
        delta = abs(pos_max - pos_min)
        BoxCenter = 0.5 * (pos_min + pos_max)

        # BoxSize is padded diff of min/max coordinates
        if self.attrs['BoxSize'] is None:
            delta *= 1.0 + self.attrs['BoxPad']
            BoxSize = numpy.ceil(delta) # round up to nearest integer
        else:
            BoxSize = self.attrs['BoxSize']

        return BoxSize, BoxCenter

    def to_mesh(self, Nmesh=None, BoxSize=None, BoxCenter=None, dtype='f4', interlaced=False,
                compensated=False, resampler='cic', fkp_weight='FKPWeight',
                comp_weight='Weight', selection='Selection',
                position='Position', bbox_from_species=None, window=None, nbar=None):

        """
        Convert the FKPCatalog to a mesh, which knows how to "paint" the
        FKP density field.

        Additional keywords to the :func:`to_mesh` function include the
        FKP weight column, completeness weight column, and the column
        specifying the number density as a function of redshift.

        Parameters
        ----------
        Nmesh : int, 3-vector, optional
            the number of cells per box side; if not specified in `attrs`, this
            must be provided
        dtype : str, dtype, optional
            the data type of the mesh when painting
        interlaced : bool, optional
            whether to use interlacing to reduce aliasing when painting the
            particles on the mesh
        compensated : bool, optional
            whether to apply a Fourier-space transfer function to account for
            the effects of the gridding + aliasing
        resampler : str, optional
            the string name of the resampler to use when interpolating the
            particles to the mesh; see ``pmesh.window.methods`` for choices
        fkp_weight : str, optional
            the name of the column in the source specifying the FKP weight;
            this weight is applied to the FKP density field:
            ``n_data - alpha*n_randoms``
        comp_weight : str, optional
            the name of the column in the source specifying the completeness
            weight; this weight is applied to the individual fields, either
            ``n_data``  or ``n_random``
        selection : str, optional
            the name of the column used to select a subset of the source when
            painting
        position : str, optional
            the name of the column that specifies the position data of the
            objects in the catalog
        bbox_from_species: str, optional
            if given, use the species to infer a bbox.
            if not give, will try random, then data (if random is empty)
        window : deprecated.
            use resampler=
        nbar: deprecated.
            deprecated. set nbar in the call to FKPCatalog()
        """
        from .catalogmesh import FKPCatalogMesh

        if window is not None:
            import warnings
            resampler = window
            warnings.warn("the window argument is deprecated. Use resampler= instead", DeprecationWarning)

        # verify that all of the required columns exist
        for name in self.species:
            for col in [fkp_weight, comp_weight]:
                if col not in self[name]:
                    raise ValueError("the '%s' species is missing the '%s' column" %(name, col))

        if Nmesh is None:
            try:
                Nmesh = self.attrs['Nmesh']
            except KeyError:
                raise ValueError("cannot convert FKP source to a mesh; 'Nmesh' keyword is not "
                                 "supplied and the FKP source does not define one in 'attrs'.")

        # first, define the Cartesian box
        if bbox_from_species is not None:
            BoxSize1, BoxCenter1 = self._define_bbox(position, selection, bbox_from_species)
        else:
            if self['randoms'].csize > 0:
                BoxSize1, BoxCenter1 = self._define_bbox(position, selection, "randoms")
            else:
                BoxSize1, BoxCenter1 = self._define_bbox(position, selection, "data")

        if BoxSize is None:
            BoxSize = BoxSize1

        if BoxCenter is None:
            BoxCenter = BoxCenter1

        # log some info
        if self.comm.rank == 0:
            self.logger.info("BoxSize = %s" %str(BoxSize))
            self.logger.info("BoxCenter = %s" %str(BoxCenter))

        # initialize the FKP mesh
        kws = {'Nmesh':Nmesh, 'BoxSize':BoxSize, 'BoxCenter' : BoxCenter, 'dtype':dtype, 'selection':selection}
        return FKPCatalogMesh(self,
                              nbar=self.nbar,
                              comp_weight=comp_weight,
                              fkp_weight=fkp_weight,
                              position=position,
                              value='Value',
                              interlaced=interlaced,
                              compensated=compensated,
                              resampler=resampler,
                              **kws)



import numpy
import os

from .base import FileType
from . import tools
from six import string_types

def getsize(filename, header_size, rowsize):
    """
    The default method to determine the size of the binary file

    The "size" is defined as the number of rows, where each
    row has of size of `rowsize` in bytes.

    Notes
    -----
    *   This assumes the input file is not compressed
    *   This function does not depend on the layout of the
        binary file, i.e., if the data is formatted in actual
        rows or not

    Raises
    ------
    ValueError :
        If the function determines a fractional number of rows

    Parameters
    ----------
    filename : str
        the name of the binary file
    header_size : int
        the size of the header in bytes, which will be skipped
        when determining the number of rows
    rowsize : int
        the size of the data in each row in bytes
    """
    bytesize = os.path.getsize(filename)
    size, remainder = divmod(bytesize-header_size, rowsize)
    if remainder != 0:
        raise ValueError("byte size mismatch -- fractional rows found")
    return size


class BinaryFile(FileType):
    """
    A file object to handle the reading of columns of data from
    a binary file.

    .. warning::

        This assumes the data is stored in a column-major format

    Parameters
    ----------
    path : str
        the name of the binary file to load
    dtype : numpy.dtype or list of tuples
        the dtypes of the columns to load; this should be either a ``numpy.dtype``
        or be able to be converted to one via a :func:`numpy.dtype` call
    offsets : dict, optional
        a dictionay specifying the byte offsets of each column in the binary
        file; if not supplied, the offsets are inferred from the dtype size
        of each column, assuming a fixed header size, and contiguous storage
    header_size : int, optional
        the size of the header in bytes
    size : int, optional
        the number of objects in the binary file; if not provided, the value
        is inferred from the dtype and the total size of the file in bytes
    """
    def __init__(self, path, dtype, offsets=None, header_size=0, size=None):

        self.path = path
        self.dataset = "*"

        # set the data type
        self.dtype = dtype
        if not isinstance(self.dtype, numpy.dtype):
            self.dtype = numpy.dtype(self.dtype)

        # determine the size (either an int or a function)
        if size is None:
            size = lambda fn: getsize(fn, header_size, self.dtype.itemsize)
        if callable(size):
            self.size = size(self.path)
        elif isinstance(size, int):
            self.size = size
        else:
            raise TypeError("`size` keyword should be a callable or integer")

        # use the input offsets dict
        if offsets is not None:
            if not isinstance(offsets, dict):
                raise TypeError("`offsets` keyword should be a dict")
            self.offsets = offsets.copy()

            # make sure each column in dtype is in the offsets table
            if not all(col in self.offsets for col in self):
                raise ValueError("missing some dtype columns in the input `offsets` dict")
        # create the dictionary of offsets
        else:
            self.offsets = {}
            for col in self:
                self.offsets[col] = self._default_byte_offset(col, header_size=header_size)

    def _default_byte_offset(self, col, header_size=0):
        """
        Internal function to return the offset in bytes
        for the column name

        This assumes consecutive storage of columns, so the offset
        for the second column is the size of the full array of the
        first column (plus header size)
        """
        offset = header_size
        cols = self.keys()
        i = 0
        while i < cols.index(col):
            offset += self.size*self.dtype[cols[i]].itemsize
            i += 1

        return offset

    def read(self, columns, start, stop, step=1):
        """
        Read the specified column(s) over the given range

        'start' and 'stop' should be between 0 and :attr:`size`,
        which is the total size of the binary file (in particles)

        Parameters
        ----------
        columns : str, list of str
            the name of the column(s) to return
        start : int
            the row integer to start reading at
        stop : int
            the row integer to stop reading at
        step : int, optional
            the step size to use when reading; default is 1

        Returns
        -------
        numpy.array
            structured array holding the requested columns over
            the specified range of rows
        """
        if isinstance(columns, string_types): columns = [columns]

        dt = [(col, self.dtype[col]) for col in columns]
        toret = numpy.empty(tools.get_slice_size(start, stop, step), dtype=dt)

        with open(self.path, 'rb') as ff:

            for col in columns:
                offset = self.offsets[col]
                dtype = self.dtype[col]
                ff.seek(offset, 0)
                ff.seek(start * dtype.itemsize, 1)
                toret[col][:] = numpy.fromfile(ff, count=stop-start, dtype=dtype)[::step]

        return toret



from nbodykit.base.mesh import MeshSource
from nbodykit import _global_options
import numpy
import logging
import warnings

# for converting from particle to mesh
from pmesh import window
from pmesh.pm import RealField, ComplexField

class CatalogMesh(MeshSource):
    """
    A mesh generated by resampling a Catalog with the given parameters.

    The original CatalogSource object is stored as the :attr:`base` attribute.

    Parameters
    ----------
    source : CatalogSource
        the input catalog that we are viewing as a mesh
    BoxSize :
        the size of the box
    Nmesh : int, 3-vector
        the number of cells per mesh side
    dtype : str
        the data type of the values stored on mesh
    Weight : array_like, Column, None
        column that specifies the weight value for each
        particle in the ``source`` to use when gridding
    Value : array_like, Column, None
        column that specifies the field value for each particle;
        the mesh stores a weighted average of this column
    Selection : array_like, Column, None
        column that selects the subset of particles to grid
        to the mesh
    Position : array_like, Column, None
        column in ``source`` specifying the position coordinates; default
        is ``Position``
    interlaced : bool, optional
        use the interlacing technique of Sefusatti et al. 2015 to reduce
        the effects of aliasing on Fourier space quantities computed
        from the mesh
    compensated : bool, optional
        whether to correct for the window introduced by the grid
        interpolation scheme
    window : str, optional
        the string specifying which window interpolation scheme to use;
        see ``pmesh.window.methods``


    'Weight', 'Value', 'Selection', 'Position' are items of the collection
    that can be reassigned after the creation of the object.

    """
    logger = logging.getLogger('CatalogMesh')

    def __repr__(self):
        return "(%s as CatalogMesh)" % repr(self.source)

    def __init__(self, source, Nmesh, BoxSize,
                    Position,
                    dtype='f4',
                    resampler='cic',
                    compensated=False,
                    interlaced=False,
                    Value=None,
                    Selection=None,
                    Weight=None,
                    **kwargs):
        from nbodykit.base.catalog import CatalogSourceBase

        assert isinstance(source, CatalogSourceBase)

        self._columns = {}

        # copy meta-data from source too
        self.attrs.update(source.attrs)

        MeshSource.__init__(self, source.comm,
                                  Nmesh,
                                  BoxSize,
                                  dtype)

        self.source = source

        # store others as straight attributes
        self.dtype = dtype

        self.Position = Position
        self.Weight = Weight
        self.Value = Value
        self.Selection = Selection

        self.attrs['interlaced'] = interlaced
        self.attrs['compensated'] = compensated
        self.attrs['resampler'] = str(resampler)

    @property
    def interlaced(self):
        """
        Whether to use interlacing when interpolating the density field.
        See :ref:`the documentation <interlacing>` for further details.

        See also: Section 3.1 of
        `Sefusatti et al. 2015 <https://arxiv.org/abs/1512.07295>`_
        """
        return self.attrs['interlaced']

    @interlaced.setter
    def interlaced(self, interlaced):
        self.attrs['interlaced'] = interlaced

    @property
    def window(self):
        return self.attrs['resampler']

    @window.setter
    def window(self, value):
        self.resampler = value

    @property
    def resampler(self):
        """
        String specifying the name of the interpolation kernel when
        gridding the density field.

        See :ref:`the documentation <resampler-kernel>` for further details.

        .. note::
            Valid values must be in :attr:`pmesh.resampler.methods`
        """
        return self.attrs['resampler']

    @resampler.setter
    def resampler(self, value):
        assert value in window.methods
        self.attrs['resampler'] = value.lower() # lower to compare with compensation

    @property
    def compensated(self):
        """
        Boolean flag to indicate whether to correct for the windowing
        kernel introduced when interpolating the discrete particles to
        a continuous field.

        See :ref:`the documentation <compensation>` for further details.
        """
        return self.attrs['compensated']

    @compensated.setter
    def compensated(self, value):
        self.attrs['compensated'] = value


    def to_real_field(self, out=None, normalize=True):
        """
        Paint the density field, by interpolating the position column
        on to the mesh.

        This computes the following meta-data attributes in the process of
        painting, returned in the :attr:`attrs` attributes of the returned
        RealField object:

        - N : int
            the (unweighted) total number of objects painted to the mesh
        - W : float
            the weighted number of total objects, equal to the collective
            sum of the 'weight' column
        - shotnoise : float
            the Poisson shot noise, equal to the volume divided by ``N``
        - num_per_cell : float
            the mean number of weighted objects per cell

        .. note::

            The density field on the mesh is normalized as :math:`1+\delta`,
            such that the collective mean of the field is unity.

        See the :ref:`documentation <painting-mesh>` on painting for more
        details on painting catalogs to a mesh.

        Returns
        -------
        real : :class:`pmesh.pm.RealField`
            the painted real field; this has a ``attrs`` dict storing meta-data
        """

        pm = self.pm
        Nlocal = 0 # (unweighted) number of particles read on local rank
        Wlocal = 0 # (weighted) number of particles read on local rank
        W2local = 0 # sum of weight square. This is used to estimate shotnoise.

        # the paint brush window
        resampler = window.methods[self.resampler]

        # initialize the RealField to return
        if out is not None:
            assert isinstance(out, RealField), "output of to_real_field must be a RealField"
            numpy.testing.assert_array_equal(out.pm.Nmesh, pm.Nmesh)
            toret = out
        else:
            toret = RealField(pm)
            toret[:] = 0

        # for interlacing, we need two empty meshes if out was provided
        # since out may have non-zero elements, messing up our interlacing sum
        if self.interlaced:

            real1 = RealField(pm)
            real1[:] = 0

            # the second, shifted mesh (always needed)
            real2 = RealField(pm)
            real2[:] = 0

        Position = self.Position
        Weight = self.Weight
        Value = self.Value
        Selection = self.Selection

        # ensure the slices are synced, since decomposition is collective
        Nlocalmax = max(pm.comm.allgather(len(Position)))

        H = pm.BoxSize / pm.Nmesh

        # paint data in chunks on each rank;
        # we do this by chunk 8 million is pretty big anyways.
        max_chunksize = _global_options['paint_chunk_size']

        # use a local scope to avoid having two copies of data in memory
        def dochunk(s):
            if len(Position) != 0:

                # selection has to be computed many times when data is `large`.

                columns = [Position[s]]
                if Weight is not None:
                    columns.append(Weight[s])
                if Value is not None:
                    columns.append(Value[s])
                if Selection is not None:
                    columns.append(Selection[s])

                # be sure to use the source to compute
                data = self.source.compute(columns)

                sel = Ellipsis if Selection is None else data.pop()
                value = None      if Value is None else data.pop()[sel]
                weight = None     if Weight is None else data.pop()[sel]
                position = data.pop()[sel]

            else:
                # workaround a potential dask issue on empty dask arrays
                position = numpy.empty((0, 3), dtype=Position.dtype)
                weight = None
                value = None

            if weight is None:
                weight = numpy.ones(len(position))

            if value is None:
                value = numpy.ones(len(position))

            # track total (selected) number and sum of weights
            Nlocal = len(position)
            Wlocal = weight.sum()
            W2local = (weight ** 2).sum()

            # no interlacing
            if not self.interlaced:
                lay = pm.decompose(position, smoothing=0.5 * resampler.support)
            else:
                lay = pm.decompose(position, smoothing=1.0 * resampler.support)

            # if we are receiving too many particles, abort and retry with a smaller chunksize
            newlengths = pm.comm.allgather(lay.newlength)
            if any([newlength > 2 * max_chunksize for newlength in newlengths]):
                if pm.comm.rank == 0:
                    self.logger.info("Throttling chunksize as some ranks will receive too many particles. (%d > %d)" % (max(newlengths), max_chunksize * 2))
                raise StopIteration

            p = lay.exchange(position)
            w = lay.exchange(weight)
            v = lay.exchange(value)

            if not self.interlaced:
                pm.paint(p, mass=w * v, resampler=resampler, hold=True, out=toret)

            # interlacing: use 2 meshes separated by 1/2 cell size
            else:
                # in mesh units
                shifted = pm.affine.shift(0.5)

                # paint to two shifted meshes
                pm.paint(p, mass=w * v, resampler=resampler, hold=True, out=real1)
                pm.paint(p, mass=w * v, resampler=resampler, transform=shifted, hold=True, out=real2)

            return Nlocal, Wlocal, W2local

        import gc
        i = 0
        chunksize = max_chunksize
        while i < Nlocalmax:

            s = slice(i, i + chunksize)

            if pm.comm.rank == 0:
                self.logger.info("Chunk %d ~ %d / %d " % (i, i + chunksize, Nlocalmax))

            try:
                Nlocal1, Wlocal1, W2local1 = dochunk(s)
                chunksize = min(max_chunksize, int(chunksize * 1.5))
            except StopIteration:
                chunksize = chunksize // 2
                if chunksize < 1:
                    raise RuntimeError("Cannot find a chunksize that fits into memory.")
                continue
            finally:
                # collect unfreed items
                gc.collect()

            Nlocal += Nlocal1
            Wlocal += Wlocal1
            W2local += W2local1

            Nglobal = pm.comm.allreduce(Nlocal)

            if pm.comm.rank == 0:
                self.logger.info("painted %d out of %d objects to mesh"
                    % (Nglobal, self.source.csize))

            i = i + chunksize

        # now the loop over particles is done

        if not self.interlaced:
            # nothing to do, toret is already filled.
            pass
        else:
            # compose the two interlaced fields into the final result.
            c1 = real1.r2c()
            c2 = real2.r2c()

            # and then combine
            for k, s1, s2 in zip(c1.slabs.x, c1.slabs, c2.slabs):
                kH = sum(k[i] * H[i] for i in range(3))
                s1[...] = s1[...] * 0.5 + s2[...] * 0.5 * numpy.exp(0.5 * 1j * kH)

            # FFT back to real-space
            # NOTE: cannot use "toret" here in case user supplied "out"
            c1.c2r(real1)

            # need to add to the returned mesh if user supplied "out"
            toret[:] += real1[:]


        # unweighted number of objects
        N = pm.comm.allreduce(Nlocal)

        # weighted number of objects
        W = pm.comm.allreduce(Wlocal)

        # weighted number of objects
        W2 = pm.comm.allreduce(W2local)

        # weighted number density (objs/cell)
        nbar = 1. * W / numpy.prod(pm.Nmesh)

        # make sure we painted something or nbar is nan; in which case
        # we set the density to uniform everywhere.
        if N == 0:
            warnings.warn(("trying to paint particle source to mesh, "
                           "but no particles were found!"),
                            RuntimeWarning
                        )

        # shot noise is volume / un-weighted number
        shotnoise = numpy.prod(pm.BoxSize) * W2 / W ** 2

        # save some meta-data
        toret.attrs = {}
        toret.attrs['shotnoise'] = shotnoise
        toret.attrs['N'] = N
        toret.attrs['W'] = W
        toret.attrs['W2'] = W
        toret.attrs['num_per_cell'] = nbar

        csum = toret.csum()
        if pm.comm.rank == 0:
            self.logger.info("painted %d out of %d objects to mesh" %(N, self.source.csize))
            self.logger.info("mean particles per cell is %g", nbar)
            self.logger.info("sum is %g ", csum)

        if normalize:
            if nbar > 0:
                toret[...] /= nbar
            else:
                toret[...] = 1

            if pm.comm.rank == 0:
                self.logger.info("normalized the convention to 1 + delta")

        return toret

    @property
    def actions(self):
        """
        The actions to apply to the interpolated density field, optionally
        included the compensation correction.
        """
        actions = MeshSource.actions.fget(self)
        if self.compensated:
            actions = self._get_compensation() + actions
        return actions

    def _get_compensation(self):
        return get_compensation(self.interlaced, self.resampler)

def get_compensation(interlaced, resampler):
    """
    Return the compensation function, which corrects for the
    windowing kernel.

    The compensation function is computed as:

    - if ``interlaced = True``:
      - :func:`CompensateCIC` if using CIC window
      - :func:`CompensateTSC` if using TSC window
      - :func:`CompensatePCS` if using PCS window
    - if ``interlaced = False``:
      - :func:`CompensateCICShotnoise` if using CIC window
      - :func:`CompensateTSCShotnoise` if using TSC window
      - :func:`CompensatePCSShotnoise` if using PCS window
    """
    if interlaced:
        d = {'cic' : CompensateCIC,
             'tsc' : CompensateTSC,
             'pcs' : CompensatePCS,
            }
    else:
        d = {'cic' : CompensateCICShotnoise,
             'tsc' : CompensateTSCShotnoise,
             'pcs' : CompensatePCSShotnoise,
            }

    if not resampler in d:
        raise ValueError("compensation for window %s is not defined" % resampler)

    filter = d[resampler]

    return [('complex', filter, "circular")]

def CompensateTSC(w, v):
    """
    Return the Fourier-space kernel that accounts for the convolution of
    the gridded field with the TSC window function in configuration space.

    .. note::
        see equation 18 (with p=3) of
        `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

    Parameters
    ----------
    w : list of arrays
        the list of "circular" coordinate arrays, ranging from
        :math:`[-\pi, \pi)`.
    v : array_like
        the field array
    """
    for i in range(3):
        wi = w[i]
        tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 3
        v = v / tmp
    return v

def CompensatePCS(w, v):
    """
    Return the Fourier-space kernel that accounts for the convolution of
    the gridded field with the PCS window function in configuration space.

    .. note::
        see equation 18 (with p=4) of
        `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

    Parameters
    ----------
    w : list of arrays
        the list of "circular" coordinate arrays, ranging from
        :math:`[-\pi, \pi)`.
    v : array_like
        the field array
    """
    for i in range(3):
        wi = w[i]
        tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 4
        v = v / tmp
    return v

def CompensateCIC(w, v):
    """
    Return the Fourier-space kernel that accounts for the convolution of
    the gridded field with the CIC window function in configuration space

    .. note::
        see equation 18 (with p=2) of
        `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

    Parameters
    ----------
    w : list of arrays
        the list of "circular" coordinate arrays, ranging from
        :math:`[-\pi, \pi)`.
    v : array_like
        the field array
    """
    for i in range(3):
        wi = w[i]
        tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 2
        tmp[wi == 0.] = 1.
        v = v / tmp
    return v

def CompensateTSCShotnoise(w, v):
    """
    Return the Fourier-space kernel that accounts for the convolution of
    the gridded field with the TSC window function in configuration space,
    as well as the approximate aliasing correction to the first order

    .. note::
        see equation 20 of
        `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

    Parameters
    ----------
    w : list of arrays
        the list of "circular" coordinate arrays, ranging from
        :math:`[-\pi, \pi)`.
    v : array_like
        the field array
    """
    for i in range(3):
        wi = w[i]
        s = numpy.sin(0.5 * wi)**2
        v = v / (1 - s + 2./15 * s**2) ** 0.5
    return v

def CompensatePCSShotnoise(w, v):
    """
    Return the Fourier-space kernel that accounts for the convolution of
    the gridded field with the PCS window function in configuration space,
    as well as the approximate aliasing correction to first order

    .. note::

        YF: I derived this by fitting the result to s ** 3
        according to the form given in equation 20 of Jing et al.
        It must be possible to derive this manually.

    Parameters
    ----------
    w : list of arrays
        the list of "circular" coordinate arrays, ranging from
        :math:`[-\pi, \pi)`.
    v : array_like
        the field array
    """
    for i in range(3):
        wi = w[i]
        s = numpy.sin(0.5 * wi)**2
        v = v / (1 - 4./3. * s + 2./5. * s**2 - 4./315. * s**3) ** 0.5
    return v

def CompensateCICShotnoise(w, v):
    """
    Return the Fourier-space kernel that accounts for the convolution of
    the gridded field with the CIC window function in configuration space,
    as well as the approximate aliasing correction to the first order

    .. note::
        see equation 20 of
        `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

    Parameters
    ----------
    w : list of arrays
        the list of "circular" coordinate arrays, ranging from
        :math:`[-\pi, \pi)`.
    v : array_like
        the field array
    """
    for i in range(3):
        wi = w[i]
        v = v / (1 - 2. / 3 * numpy.sin(0.5 * wi) ** 2) ** 0.5
    return v




from nbodykit.base.mesh import MeshFilter

import numpy

class TopHat(MeshFilter):
    """ A TopHat filter defined in Fourier space.

        Notes
        -----
        A fourier space filter is different from a configuration space
        filter. The TopHat in fourier space creates ringing effects
        due to the truncation / discretization of modes.

    """
    kind = 'wavenumber'
    mode = 'complex'

    def __init__(self, r):
        """
            Parameters
            ----------
            r : float
                radius of the TopHat filter
        """
        self.r = r

    def filter(self, k, v):
        r = self.r
        k = sum(ki ** 2 for ki in k) ** 0.5
        kr = k * r
        w = 3 * (numpy.sin(kr) / kr **3 - numpy.cos(kr) / kr ** 2)
        w[k == 0] = 1.0
        return w * v

class Gaussian(MeshFilter):
    """ A gaussian filter

        .. math ::

            G(r) = exp(-0.5 k^2 r^2)

    """
    kind = 'wavenumber'
    mode = 'complex'

    def __init__(self, r):
        """
            Parameters
            ----------
            r : float
                radius of the Gaussian filter
        """
        self.r = r

    def filter(self, k, v):
        r = self.r
        k2 = sum(ki ** 2 for ki in k)
        return numpy.exp(- 0.5 * k2 * r**2)




import numpy
import dask.array as da
from six import string_types
from nbodykit.utils import deprecate
from nbodykit import _global_options
def StackColumns(*cols):
    """
    Stack the input dask arrays vertically, column by column.

    This uses :func:`dask.array.vstack`.

    Parameters
    ----------
    *cols : :class:`dask.array.Array`
        the dask arrays to stack vertically together

    Returns
    -------
    :class:`dask.array.Array` :
        the dask array where columns correspond to the input arrays

    Raises
    ------
    TypeError
        If the input columns are not dask arrays
    """
    cols = da.broadcast_arrays(*cols)
    return da.vstack(cols).T

def ConcatenateSources(*sources, **kwargs):
    """
    Concatenate CatalogSource objects together, optionally including only
    certain columns in the returned source.

    .. note::
        The returned catalog object carries the meta-data from only
        the first catalog supplied to this function (in the ``attrs`` dict).

    Parameters
    ----------
    *sources : subclass of :class:`~nbodykit.base.catalog.CatalogSource`
        the catalog source objects to concatenate together
    columns : str, list of str, optional
        the columns to include in the concatenated catalog

    Returns
    -------
    CatalogSource :
        the concatenated catalog source object

    Examples
    --------
    >>> from nbodykit.lab import *
    >>> source1 = UniformCatalog(nbar=100, BoxSize=1.0)
    >>> source2 = UniformCatalog(nbar=100, BoxSize=1.0)
    >>> print(source1.csize, source2.csize)
    >>> combined = transform.ConcatenateSources(source1, source2, columns=['Position', 'Velocity'])
    >>> print(combined.csize)
    """
    from nbodykit.base.catalog import CatalogSource

    columns = kwargs.get('columns', None)
    if isinstance(columns, string_types):
        columns = [columns]

    # concatenate all columns, if none provided
    if columns is None or columns == []:
        columns = sources[0].columns

    # check comms
    if not all(src.comm == sources[0].comm for src in sources):
        raise ValueError("cannot concatenate sources: comm mismatch")

    # check all columns are there
    for source in sources:
        if not all(col in source for col in columns):
            raise ValueError(("cannot concatenate sources: columns are missing "
                              "from some sources"))
    # the total size
    size = numpy.sum([src.size for src in sources], dtype='intp')

    data = {}
    for col in columns:
        data[col] = da.concatenate([src[col] for src in sources], axis=0)

    toret = CatalogSource._from_columns(size, sources[0].comm, **data)
    toret.attrs.update(sources[0].attrs)
    return toret

def ConstantArray(value, size, chunks=100000):
    """
    Return a dask array of the specified ``size`` holding a single value.

    This uses numpy's "stride tricks" to avoid replicating
    the data in memory for each element of the array.

    Parameters
    ----------
    value : float
        the scalar value to fill the array with
    size : int
        the length of the returned dask array
    chunks : int, optional
        the size of the dask array chunks
    """
    ele = numpy.array(value)
    toret = numpy.lib.stride_tricks.as_strided(ele, [size] + list(ele.shape), [0] + list(ele.strides))
    return da.from_array(toret, chunks=chunks, name=False)


def CartesianToEquatorial(pos, observer=[0,0,0], frame='icrs'):
    """
    Convert Cartesian position coordinates to equatorial right ascension
    and declination, using the specified observer location.

    .. note::
        RA and DEC will be returned in degrees, with RA in the range [0,360]
        and DEC in the range [-90, 90].

    Parameters
    ----------
    pos : array_like
        a N x 3 array holding the Cartesian position coordinates
    observer : array_like
        a length 3 array holding the observer location
    frame : string
        A string, 'icrs' or 'galactic'. The frame of the input position.
        Use 'icrs' if the cartesian position is already in Equatorial.

    Returns
    -------
    ra, dec : array_like
        the right ascension and declination coordinates, in degrees. RA
        will be in the range [0,360] and DEC in the range [-90, 90]
    """

    # split x, y, z to signify that we do not need to have pos
    # as a full chunk in the last dimension.
    # this is useful when we use apply_gufunc.

    x, y, z = [pos[..., i] - observer[i] for i in range(3)]

    if frame == 'icrs':
        # FIXME: Convert these to a gufunc that uses astropy?
        # might be a step backward.

        # from equatorial to equatorial
        s = da.hypot(x, y)
        lon = da.arctan2(y, x)
        lat = da.arctan2(z, s)

        # convert to degrees
        lon = da.rad2deg(lon)
        lat = da.rad2deg(lat)
        # wrap lon to [0,360]
        lon = da.mod(lon-360., 360.)
        ra, dec = lon, lat
    else:
        from astropy.coordinates import SkyCoord

        def cart_to_eq(x, y, z):
            try:
                sc = SkyCoord(x, y, z, representation_type='cartesian', frame=frame)
                scg = sc.transform_to(frame='icrs')
                scg.representation_type = 'unitspherical'
            except:
                sc = SkyCoord(x, y, z, representation='cartesian', frame=frame)
                scg = sc.transform_to(frame='icrs')
                scg.representation = 'unitspherical'

            ra, dec = scg.ra.value, scg.dec.value

            return ra, dec

        dtype = pos.dtype
        ra, dec = da.apply_gufunc(cart_to_eq, '(),(),()->(),()', x, y, z, output_dtypes=[dtype, dtype])

    return da.stack((ra, dec), axis=0)

def CartesianToSky(pos, cosmo, velocity=None, observer=[0,0,0], zmax=100., frame='icrs'):
    r"""
    Convert Cartesian position coordinates to RA/Dec and redshift,
    using the specified cosmology to convert radial distances from
    the origin into redshift.

    If velocity is supplied, the returned redshift accounts for the
    additional peculiar velocity shift.

    Users should ensure that ``zmax`` is larger than the largest possible
    redshift being considered to avoid an interpolation exception.

    .. note::
        Cartesian coordinates should be in units of Mpc/h and velocity
        should be in units of km/s.

    Parameters
    ----------
    pos : dask array
        a N x 3 array holding the Cartesian position coordinates in Mpc/h
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`
        the cosmology used to meausre the comoving distance from ``redshift``
    velocity : array_like
        a N x 3 array holding velocity in km/s
    observer : array_like, optional
        a length 3 array holding the observer location
    zmax : float, optional
        the maximum possible redshift, should be set to a reasonably large
        value to avoid interpolation failure going from comoving distance
        to redshift
    frame : string ('icrs' or 'galactic')
        speciefies which frame the Cartesian coordinates is. Useful if you know
        the simulation (usually cartesian) is in galactic units but you want
        to convert to the icrs (ra, dec) usually used in surveys.

    Returns
    -------
    ra, dec, z : dask array
        the right ascension (in degrees), declination (in degrees), and
        redshift coordinates. RA will be in the range [0,360] and DEC in the
        range [-90, 90]

    Notes
    -----
    If velocity is provided, redshift-space distortions are added to the
    real-space redshift :math:`z_\mathrm{real}`, via:

    .. math::

            z_\mathrm{redshift} = ( v_\mathrm{pec} / c ) (1 + z_\mathrm{reals})

    Raises
    ------
    TypeError
        If the input columns are not dask arrays
    """
    from astropy.constants import c
    from scipy.interpolate import interp1d

    if not isinstance(pos, da.Array):
        pos = da.from_array(pos, chunks=100000)

    pos = pos - observer
    # RA,dec coordinates (in degrees)
    ra, dec = CartesianToEquatorial(pos, frame=frame)

    # the distance from the origin
    r = da.linalg.norm(pos, axis=-1)

    def z_from_comoving_distance(x):
        zgrid = numpy.logspace(-8, numpy.log10(zmax), 1024)
        zgrid = numpy.concatenate([[0.], zgrid])
        rgrid = cosmo.comoving_distance(zgrid)
        return interp1d(rgrid, zgrid)(x)

    # invert distance - redshift relation
    z = r.map_blocks(z_from_comoving_distance)

    # add in velocity offsets?
    if velocity is not None:

        vpec = (pos * velocity).sum(axis=-1) / r

        z += vpec / c.to('km/s').value * (1 + z)

    return da.stack((ra, dec, z), axis=0)

def SkyToUnitSphere(ra, dec, degrees=True, frame='icrs'):
    """
    Convert sky coordinates (``ra``, ``dec``) to Cartesian coordinates on
    the unit sphere.

    Parameters
    ----------
    ra : :class:`dask.array.Array`; shape: (N,)
        the right ascension angular coordinate
    dec : :class:`dask.array.Array`; ; shape: (N,)
        the declination angular coordinate
    degrees : bool, optional
        specifies whether ``ra`` and ``dec`` are in degrees or radians
    frame : string ('icrs' or 'galactic')
        speciefies which frame the Cartesian coordinates is. Useful if you know
        the simulation (usually cartesian) is in galactic units but you want
        to convert to the icrs (ra, dec) usually used in surveys.

    Returns
    -------
    pos : :class:`dask.array.Array`; shape: (N,3)
        the cartesian position coordinates, where columns represent
        ``x``, ``y``, and ``z``

    Raises
    ------
    TypeError
        If the input columns are not dask arrays
    """
    ra, dec = da.broadcast_arrays(ra, dec)

    if frame == 'icrs':
        # no frame transformation
        # put into radians from degrees
        if degrees:
            ra  = da.deg2rad(ra)
            dec = da.deg2rad(dec)

        # cartesian coordinates
        x = da.cos( dec ) * da.cos( ra )
        y = da.cos( dec ) * da.sin( ra )
        z = da.sin( dec )
        return da.vstack([x,y,z]).T
    else:
        from astropy.coordinates import SkyCoord

        if degrees:
            ra  = da.deg2rad(ra)
            dec = da.deg2rad(dec)

        def eq_to_cart(ra, dec):
            try:
                sc = SkyCoord(ra, dec, unit='rad', representation_type='unitspherical', frame='icrs')
            except:
                sc = SkyCoord(ra, dec, unit='rad', representation='unitspherical', frame='icrs')

            scg = sc.transform_to(frame=frame)
            scg = scg.cartesian

            x, y, z = scg.x.value, scg.y.value, scg.z.value
            return numpy.stack([x, y, z], axis=1)

        arr = da.apply_gufunc(eq_to_cart, '(),()->(p)', ra, dec, output_dtypes=[ra.dtype], output_sizes={'p': 3})
        return arr

def SkyToCartesian(ra, dec, redshift, cosmo, observer=[0, 0, 0], degrees=True, frame='icrs'):
    """
    Convert sky coordinates (``ra``, ``dec``, ``redshift``) to a
    Cartesian ``Position`` column.

    .. warning::

        The returned Cartesian position is in units of Mpc/h.

    Parameters
    -----------
    ra : :class:`dask.array.Array`; shape: (N,)
        the right ascension angular coordinate
    dec : :class:`dask.array.Array`; shape: (N,)
        the declination angular coordinate
    redshift : :class:`dask.array.Array`; shape: (N,)
        the redshift coordinate
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`
        the cosmology used to meausre the comoving distance from ``redshift``
    degrees : bool, optional
        specifies whether ``ra`` and ``dec`` are in degrees
    frame : string ('icrs' or 'galactic')
        speciefies which frame the Cartesian coordinates is. 

    Returns
    -------
    pos : :class:`dask.array.Array`; shape: (N,3)
        the cartesian position coordinates, where columns represent
        ``x``, ``y``, and ``z`` in units of Mpc/h

    Raises
    ------
    TypeError
        If the input columns are not dask arrays
    """
    ra, dec, redshift = da.broadcast_arrays(ra, dec, redshift)

    # pos on the unit sphere
    pos = SkyToUnitSphere(ra, dec, degrees=degrees, frame=frame)

    # multiply by the comoving distance in Mpc/h
    r = redshift.map_blocks(lambda z: cosmo.comoving_distance(z), dtype=redshift.dtype)

    return r[:,None] * pos + observer

def HaloConcentration(mass, cosmo, redshift, mdef='vir'):
    """
    Return halo concentration from halo mass, based on the analytic fitting
    formulas presented in
    `Dutton and Maccio 2014 <https://arxiv.org/abs/1402.7073>`_.

    .. note::
        The units of the input mass are assumed to be :math:`M_{\odot}/h`

    Parameters
    ----------
    mass : array_like
        either a numpy or dask array specifying the halo mass; units
        assumed to be :math:`M_{\odot}/h`
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`
        the cosmology instance used in the analytic formula
    redshift : float
        compute the c(M) relation at this redshift
    mdef : str, optional
        string specifying the halo mass definition to use; should be
        'vir' or 'XXXc' or 'XXXm' where 'XXX' is an int specifying the
        overdensity

    Returns
    -------
    concen : :class:`dask.array.Array`
        a dask array holding the analytic concentration values

    References
    ----------
    Dutton and Maccio, "Cold dark matter haloes in the Planck era: evolution
    of structural parameters for Einasto and NFW profiles", 2014, arxiv:1402.7073

    """
    from halotools.empirical_models import NFWProfile

    mass, redshift = da.broadcast_arrays(mass, redshift)

    kws = {'cosmology':cosmo.to_astropy(), 'conc_mass_model':'dutton_maccio14', 'mdef':mdef}

    def get_nfw_conc(mass, redshift):
        kw1 = {}
        kw1.update(kws)
        kw1['redshift'] = redshift
        model = NFWProfile(**kw1)
        return model.conc_NFWmodel(prim_haloprop=mass)

    return da.map_blocks(get_nfw_conc, mass, redshift, dtype=mass.dtype)

def HaloVelocityDispersion(mass, cosmo, redshift, mdef='vir'):
    """ Compute the velocity dispersion of halo from Mass.

        This is a simple model suggested by Martin White.

        See http://adsabs.harvard.edu/abs/2008ApJ...672..122E
    """

    mass, redshift = da.broadcast_arrays(mass, redshift)
    def compute_vdisp(mass, redshift):
        h = cosmo.efunc(redshift)
        return 1100. * (h * mass / 1e15) ** 0.33333

    return da.map_blocks(compute_vdisp, mass, redshift, dtype=mass.dtype)

def HaloRadius(mass, cosmo, redshift, mdef='vir'):
    r"""
    Return proper halo radius from halo mass, based on the specified mass definition.
    This is independent of halo profile, and simply returns

    .. math::

        R = \left [ 3 M /(4\pi\Delta) \right]^{1/3}

    where :math:`\Delta` is the density threshold, which depends on cosmology,
    redshift, and mass definition

    .. note::
        The units of the input mass are assumed to be :math:`M_{\odot}/h`

    Parameters
    ----------
    mass : array_like
        either a numpy or dask array specifying the halo mass; units
        assumed to be :math:`M_{\odot}/h`
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`
        the cosmology instance
    redshift : float
        compute the density threshold which determines the R(M) relation
        at this redshift
    mdef : str, optional
        string specifying the halo mass definition to use; should be
        'vir' or 'XXXc' or 'XXXm' where 'XXX' is an int specifying the
        overdensity

    Returns
    -------
    radius : :class:`dask.array.Array`
        a dask array holding the halo radius in 'physical Mpc/h [sic]'.
        This is proper Mpc/h, to convert to comoving, divide this by scaling factor.

    """
    from halotools.empirical_models import halo_mass_to_halo_radius

    mass, redshift = da.broadcast_arrays(mass, redshift)

    kws = {'cosmology':cosmo.to_astropy(), 'mdef':mdef}

    def mass_to_radius(mass, redshift):
        return halo_mass_to_halo_radius(mass=mass, redshift=redshift, **kws)

    return da.map_blocks(mass_to_radius, mass, redshift, dtype=mass.dtype)

# deprecated functions
vstack = deprecate("nbodykit.transform.vstack", StackColumns, "nbodykit.transform.StackColumns")
concatenate = deprecate("nbodykit.transform.concatenate", ConcatenateSources, "nbodykit.transform.ConcatenateSources")
SkyToCartesion = deprecate("nbodykit.transform.SkyToCartesion", SkyToCartesian, "nbodykit.transform.SkyToCartesian")



import os
import numpy
import logging
import warnings

from nbodykit import CurrentMPIComm
from nbodykit.base.mesh import MeshSource
from nbodykit.base.catalog import CatalogSource
from nbodykit import _global_options

class FFTRecon(MeshSource):
    """
    FFT based Lagrangian reconstruction algorithm in a periodic box.

    References:

        Eisenstein et al, 2007
        http://adsabs.harvard.edu/abs/2007ApJ...664..675E
        Section 3, paragraph starting with 'Restoring in full the ...'

    We follow a cleaner description in Schmitfull et al 2015,

        http://adsabs.harvard.edu/cgi-bin/bib_query?arXiv:1508.06972

    Table I, and text below. Schemes are LGS, LF2 and LRR.

    A slight difference against the paper is that Redshift distortion
    and bias are corrected in the linear order. The Random shifting
    followed Martin White's suggestion to exclude the RSD by default.
    (with default `revert_rsd_random=False`.)

    Parameters
    ----------
    data : CatalogSource,
        the data catalog, e.g. halos. `data.attrs['BoxSize']` is used if argument `BoxSize` is not given.
    ran  :  CatalogSource
        the random catalog, e.g. from a `UniformCatalog` object.
    Nmesh : int
        The size of the FFT Mesh. Rule of thumb is that the size of a mesh cell
        shall be 2 ~ 4 times smaller than the smoothing length, `R`.
    revert_rsd_random : boolean
        Revert the rsd for randoms as well as data. There are two conventions.
        either reverting rsd displacement in data displacement only(False) or
        in both data and randoms (True). Default is False.
    R : float
        The radius of smoothing. 10 to 20 Mpc/h is usually cool.
    bias : float
        The bias of the data catalog.
    f: float
        The growth rate; if non-zero, correct for RSD
    los : list 
        The direction of the line of sight for RSD. Usually (default) [0, 0, 1].
    position: string
        column to use for picking up the Position of the objects.
    BoxSize : float or array_like
        the size of the periodic box, default is to infer from the data.

    scheme : string
        The reconstruction scheme.
        `LGS` is the standard reconstruction (Lagrangian growth shift).
        `LF2` is the F2 Lagrangian reconstruction.
        `LRR` is the random-random Lagrangian reconstruction.
    """

    def __init__(self,
            data,
            ran,
            Nmesh,
            bias=1.0,
            f = 0.0,
            los = [0, 0, 1],
            R=20,
            position='Position',
            revert_rsd_random=False,
            scheme='LGS',
            BoxSize=None):

        assert scheme in ['LGS', 'LF2', 'LRR']

        assert isinstance(data, CatalogSource)
        assert isinstance(ran, CatalogSource)

        comm = data.comm

        assert data.comm == ran.comm

        from pmesh.pm import ParticleMesh

        if Nmesh is None:
            Nmesh = data.attrs['Nmesh']
        _Nmesh = numpy.empty(3, dtype='i8')
        _Nmesh[...] = Nmesh

        if BoxSize is None:
            BoxSize = data.attrs['BoxSize']

        los = numpy.array(los, dtype='f8', copy=True)
        los /= (los ** 2).sum()

        assert len(los) == 3
        assert (~numpy.isnan(los)).all()

        pm = ParticleMesh(BoxSize=BoxSize, Nmesh=_Nmesh, comm=comm)
        self.pm = pm

        if (self.pm.BoxSize / self.pm.Nmesh).max() > R:
            if comm.rank == 0:
                warnings.warn("The smoothing radius smaller than the mesh cell size. This may produce undesired numerical results.")

        assert position in data.columns
        assert position in ran.columns

        self.position = position

        MeshSource.__init__(self, comm, pm.Nmesh.copy(), pm.BoxSize.copy(), pm.dtype)

        self.attrs['bias'] = bias
        self.attrs['f'] = f
        self.attrs['los'] = los
        self.attrs['R'] = R
        self.attrs['scheme'] = scheme
        self.attrs['revert_rsd_random'] = bool(revert_rsd_random)

        self.data = data
        self.ran = ran

        if self.comm.rank == 0:
            self.logger.info("Reconstruction for bias=%g, f=%g, smoothing R=%g los=%s" % (self.attrs['bias'], self.attrs['f'], self.attrs['R'], str(self.attrs['los'])))
            self.logger.info("Reconstruction scheme = %s" % (self.attrs['scheme']))

    def to_real_field(self):
        return self.run()

    def run(self):

        s_d, s_r = self._compute_s()
        return self._helper_paint(s_d, s_r)

    def work_with(self, cat, s):
        pm = self.pm
        delta = pm.create(mode='real', value=0)

        # ensure the slices are synced, since decomposition is collective
        Nlocalmax = max(pm.comm.allgather(cat.size))

        # python 2.7 wants floats.
        nbar = (1.0 * cat.csize / self.pm.Nmesh.prod())

        chunksize = _global_options['paint_chunk_size']

        for i in range(0, Nlocalmax, chunksize):
            sl = slice(i, i + chunksize)

            if s is not None:
                dpos = (cat[self.position].astype('f4')[sl] - s[sl]).compute()
            else:
                dpos = (cat[self.position].astype('f4')[sl]).compute()

            layout = self.pm.decompose(dpos)
            self.pm.paint(dpos, layout=layout, out=delta, hold=True)

        delta[...] /= nbar

        return delta

    def _summary_field(self, field, name):
        cmean = field.cmean()
        if self.comm.rank == 0:
            self.logger.info("painted %s, mean=%g" % (name, cmean))


    def _helper_paint(self, s_d, s_r):
        """ Convert the displacements of data and random to a single reconstruction mesh object. """

        def LGS(delta_s_r):
            delta_s_d = self.work_with(self.data, s_d)
            self._summary_field(delta_s_d, "delta_s_d (shifted)")

            delta_s_d[...] -= delta_s_r
            return delta_s_d

        def LRR(delta_s_r):
            delta_s_nr = self.work_with(self.ran, -s_r)
            self._summary_field(delta_s_nr, "delta_s_nr (reverse shifted)")

            delta_d = self.work_with(self.data, None)
            self._summary_field(delta_d, "delta_d (unshifted)")

            delta_s_nr[...] += delta_s_r[...]
            delta_s_nr[...] *= 0.5
            delta_d[...] -= delta_s_nr
            return delta_d

        def LF2(delta_s_r):
            lgs = LGS(delta_s_r)
            lrr = LRR(delta_s_r) 
            lgs[...] *= 3.0 / 7.0
            lrr[...] *= 4.0 / 7.0
            lgs[...] += lrr
            return lgs

        delta_s_r = self.work_with(self.ran, s_r)
        self._summary_field(delta_s_r, "delta_s_r (shifted)")

        if self.attrs['scheme'] == 'LGS':
            delta_recon = LGS(delta_s_r)
        elif self.attrs['scheme'] == 'LF2':
            delta_recon = LF2(delta_s_r)
        elif self.attrs['scheme'] == 'LRR':
            delta_recon = LRR(delta_s_r)
            
        self._summary_field(delta_recon, "delta_recon")

        # FIXME: perhaps change to 1 + delta for consistency. But it means loss of precision in f4 
        return delta_recon

    def _compute_s(self):
        """ Computing the reconstruction displacement of data and random """

        def kernel(d):
            def kernel(k, v):
                k2 = sum(ki**2 for ki in k)
                k2[k2 == 0] = 1.0

                # reverting rsd.
                mu = sum(k[i] * self.attrs['los'][i] for i in range(len(k))) / k2 ** 0.5

                v = v * numpy.exp(-0.5 * k2 * self.attrs['R']**2)

                frac = self.attrs['bias'] * ( 1 + self.attrs['f'] / self.attrs['bias'] * mu ** 2)

                v = v / frac

                return 1j * k[d] / k2 * v
            return kernel

        delta_d = self.work_with(self.data, None)
        self._summary_field(delta_d, "delta_d (unshifted)")
        delta_d = delta_d.r2c(out=Ellipsis)

        def solve_displacement(cat, delta_d):
            dpos = cat[self.position].astype('f4').compute()
            layout = self.pm.decompose(dpos)
            s_d = numpy.zeros_like(dpos, dtype='f4')

            for d in range(3):
                delta_d.apply(kernel(d)).c2r(out=Ellipsis) \
                       .readout(dpos, layout=layout, out=s_d[..., d])
            return s_d

        s_d = solve_displacement(self.data, delta_d)
        s_d_std = (self.comm.allreduce((s_d ** 2).sum(axis=0)) / self.data.csize) ** 0.5
        if self.comm.rank == 0:
            self.logger.info("Solved displacements of data, std(s_d) = %s" % str(s_d_std))

        s_r = solve_displacement(self.ran, delta_d)
        s_r_std = (self.comm.allreduce((s_r ** 2).sum(axis=0)) / self.ran.csize) ** 0.5
        if self.comm.rank == 0:
            self.logger.info("Solved displacements of randoms, std(s_r) = %s" % str(s_r_std))

        # convention 1: shifting data only
        s_d[...] *= (1 + self.attrs['los'] * self.attrs['f'])

        # convention 2: shifting data only
        if self.attrs['revert_rsd_random']:
            s_r[...] *= (1 + self.attrs['los'] * self.attrs['f'])
        
        return s_d, s_r




from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import pytest

# debug logging
setup_logging("debug")

@MPITest([1])
def test_fftrecon(comm):
    cosmo = cosmology.Planck15
    # this should generate 15 particles
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')

    # very high bias to increase the accuracy of reconstruction;
    # since shotnoise is high.
    data = LogNormalCatalog(Plin=Plin, bias=4, nbar=1e-4, BoxSize=1024., Nmesh=64, seed=42, comm=comm)
    ran = UniformCatalog(nbar=1e-4, BoxSize=1024., seed=42, comm=comm)

    # lognormal mocks don't have the correct small scale power for reconstruction,
    # so we heavily smooth and assert the reconstruction
    # doesn't mess it all up.

    mesh = FFTRecon(data=data, ran=ran, bias=4, Nmesh=64, R=40)

    r1 = FFTPower(mesh, mode='1d')
    r2 = FFTPower(data, mode='1d')

    # reconstruction shouldn't have matter much on large scale.
    assert_allclose(r1.power['power'][:5], r2.power['power'][:5], rtol=0.05)




import numpy as np
from scipy.integrate import odeint

class Perturbation:
    """
    Perturbation Growth coefficients at several orders.

    2-LPT is implemented. This implements the single fluid
    model of Boltamann equations. Therefore it is accurate
    only in a matter dominated universe.

    All derivatives are against ``lna``.

    .. note::
        Formulas are derived from Yin Li's notes on 2LPT.

    """

    def __init__(self, a, a_normalize=1.0):

        if a is None:
            lna = np.log(np.logspace(-7, 0, 1024*10, endpoint=True))
        else:
            a = np.array(a, copy=True).ravel() # ensure this is 1-d
            if a_normalize not in a: # ensure redshift 0 is in the list, for normalization
                a = np.concatenate([[a_normalize], a])
            a.sort()
            if a[0] > 1e-7: # add a high redshift starting point.
                a = np.concatenate([[1e-7], a])
            lna = np.log(a)

        self.lna = lna

        y0 = self.get_initial_condition()

        self._D1, self._D2 = self._solve(y0, a_normalize)


    def D1(self, a, order=0):
        """ Linear order growth function.

            Parameters
            ----------
            a : float, array_like
                scaling factor
            order : int
                order of differentation; 1 for first derivative against log a.

            Returns
            -------
            array_like : linear order growth function.
        """
        lna = np.log(a)
        return np.interp(lna, self.lna, self._D1[:, order])

    def D2(self, a, order=0):
        """ Second order growth function.

            Parameters
            ----------
            a : float, array_like
                scaling factor
            order : int
                order of differentation; 1 for first derivative against log a.

            Returns
            -------
            array_like : second order growth function.
        """
        lna = np.log(a)
        return np.interp(lna, self.lna, self._D2[:, order])

    def f1(self, a):
        """ Linear order growth rate

            Parameters
            ----------
            a : float, array_like
                scaling factor
            order : int
                order of differentation; 1 for first derivative against log a.

            Returns
            -------
            array_like : linear order growth rate.
        """
        return self.D1(a, order=1) / self.D1(a, order=0)

    def f2(self, a):
        """ Second order growth rate.

            Parameters
            ----------
            a : float, array_like
                scaling factor
            order : int
                order of differentation; 1 for first derivative against log a.

            Returns
            -------
            array_like : second order growth rate.
        """
        return self.D2(a, order=1) / self.D2(a, order=0)

    def Gp(self, a):
        """
        FastPM growth factor function, eq, 19
        """
        return self.D1(a)

    def Gp2(self, a):
        """ Gp for second order LPT
        FastPM growth factor function, eq, 19
         """
        return self.D2(a)

    def gp(self, a):
        """
        Notice the derivative of D1 is against ln a but gp is d D1 / da, so
        gp = D1(a, order=1) / a
        """
        return self.D1(a, order=1) / a


    def gp2(self, a):
        """ gp for second order LPT
        """
        return self.D2(a, order=1) / a

    def Gf(self, a):
        """
        FastPM growth factor function, eq, 20
        """
        return self.D1(a, 1) * a ** 2 * self.E(a)

    def Gf2(self, a):
        """ Gf but for second order LPT
        """
        return self.D2(a, 1) * a ** 2 * self.E(a)

    def gf(self, a):
        """
        Similarly, the derivative is against ln a, so
        gf = Gf(a, order=1) / a
        """
        return 1 / a * (
            self.D1(a, 2) * a ** 2 * self.E(a) \
            +  self.D1(a, 1) * (
                    a ** 2 * self.E(a, order=1)
                +   2 * a ** 2 * self.E(a))
            )
    def gf2(self, a):
        """
            gf but for second order LPT
        """
        return 1 / a * (
            self.D2(a, 2) * a ** 2 * self.E(a) \
            +  self.D2(a, 1) * (
                    a ** 2 * self.E(a, order=1)
                +   2 * a ** 2 * self.E(a))
            )

    def E(self, a, order=0):
        """ Hubble function and derivatives against log a.

        """
        if order == 0:
            return self.efunc(a)
        else:
            return self.efunc_prime(a) * a

    def Hfac(self, a):
        return -2. - self.E(a, order=1) / self.E(a)

    def ode(self, y, lna):
        D1, F1, D2, F2 = y
        a = np.exp(lna)
        hfac = self.Hfac(a)
        omega = self.Om(a)
        F1p = hfac * F1 + 1.5 * omega * D1
        D1p = F1
        F2p = hfac * F2 + 1.5 * omega * D2 - 1.5 * omega * D1 ** 2
        D2p = F2
        return D1p, F1p, D2p, F2p

    def _solve(self, y0, a_normalize):
        # solve with critical point at a=1.0, lna=0.
        y = odeint(self.ode, y0, self.lna, tcrit=[0.], atol=0)

        v1 = []
        v2 = []
        for yi, lnai in zip(y, self.lna):
            D1, F1, D2, F2 = yi
            D1p, F1p, D2p, F2p = self.ode(yi, lnai)
            v1.append((D1, F1, F1p))
            v2.append((D2, F2, F2p))

        v1 = np.array(v1)
        v2 = np.array(v2)

        ind = abs(self.lna - np.log(a_normalize)).argmin()
        # normalization to 1 at a=a_normalize
        v1 /= v1[ind][0]
        v2 /= v2[ind][0]
        return v1, v2

class MatterDominated(Perturbation):
    """

    Perturbation with matter dominated initial condition.

    This is usually referred to the single fluid approximation as well.

    The result here is accurate upto numerical precision. If Omega0_m 

    Parameters
    ----------
    Omega0_m :
        matter density at redshift 0

    Omega0_lambda :
        Lambda density at redshift 0, default None; set to ensure flat universe.

    Omega0_k:
        Curvature density at redshift 0, default : 0

    a : array_like
        a list of time steps where the factors are exact.
        other a values are interpolated.
    """
    def __init__(self, Omega0_m, Omega0_lambda=None, Omega0_k=0, a=None, a_normalize=1.0):
        if Omega0_lambda is None:
            Omega0_lambda = 1 - Omega0_k - Omega0_m

        self.Omega0_lambda = Omega0_lambda
        self.Omega0_m = Omega0_m
        self.Omega0_k = Omega0_k
        # Om0 is added for backward compatiblity
        self.Om0 = Omega0_m
        Perturbation.__init__(self, a, a_normalize)

    def get_initial_condition(self):
        a0 = np.exp(self.lna[0])

        # matter dominated initial conditions
        y0 = [a0, a0, -3./7 * a0**2, -6. / 7 *a0**2]
        return y0

    def efunc(self, a):
        return (self.Omega0_m / a**3+ self.Omega0_k / a**2 + self.Omega0_lambda) ** 0.5

    def efunc_prime(self, a):
        return 0.5 / self.efunc(a) * (-3 * self.Omega0_m / (a**4) + -2 * self.Omega0_k / (a**3))

    def Om(self, a):
        return (self.Omega0_m / a ** 3) / self.efunc(a) ** 2

class RadiationDominated(Perturbation):
    """

    Perturbation with Radiation dominated initial condition

    This is approximated because the single fluid scale independent
    solution will need an initial condition that comes from a
    true Boltzmann code.

    Here, the first order result is tuned to agree
    at sub-percent level comparing to a true multi-fluid
    boltzmann code under Planck15 cosmology.

    Parameters
    ----------
    cosmo: :class:`~nbodykit.cosmology.core.Cosmology`
        a astropy Cosmology like object.

    a : array_like
        a list of time steps where the factors are exact.
        other a values are interpolated.
    """
    def __init__(self, cosmo, a=None, a_normalize=1.0):
#        assert cosmo.Ogamma0 == 0
#        assert cosmo.Onu0 == 0

        self._cosmo = cosmo
        self.Omega0_m = cosmo.Om0

        # Om0 is added for backward compatiblity
        self.Om0 = self.Omega0_m
        self.Omega0_gamma = cosmo.Ogamma0

        Perturbation.__init__(self, a, a_normalize)

    def get_initial_condition(self):
        # suggested by Yin Li, not so sensitive to a_H
        # as long as it is early enough.
        # I used this number to ensure best agreement with CLASS
        # for Planck15. (1e-4 rtol)
        a_H = 5.22281250e-05

        a0 = np.exp(self.lna[0])
        Om = self.Om(a0)

        D1i = np.log(a0 / a_H)
        # radiation dominated initial conditions
        y0 = [
            D1i,
            1.0,
             -1.5 * Om * (D1i ** 2 - 4 * D1i + 6),
             -1.5 * Om * (D1i ** 2 - 2 * D1i + 2)]
        return y0

    def efunc(self, a):
        z = 1. / a - 1.0
        return self._cosmo.efunc(z)

    def efunc_prime(self, a):
        z = 1. / a - 1.0
        return self._cosmo.efunc_prime(z)

    def Om(self, a):
        z = 1./a-1
        return self._cosmo.Omega_b(z) + self._cosmo.Omega_cdm(z) # non-relativistic 


from nbodykit.utils import deprecate

PerturbationGrowth = deprecate("PerturbationGrowth", RadiationDominated)




from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_array_equal, assert_allclose
import kdcount.correlate as correlate
import os
import pytest

setup_logging()

def make_corrfunc_input(data, cosmo):
    ra = gather_data(data, 'RA')
    dec = gather_data(data, 'DEC')
    z = gather_data(data, 'Redshift')
    rdist = cosmo.comoving_distance(z)
    return numpy.vstack([ra, dec, rdist]).T

def gather_data(source, name):
    return numpy.concatenate(source.comm.allgather(source[name].compute()), axis=0)

def generate_sim_data(seed, comm):
    return UniformCatalog(nbar=3e-4, BoxSize=512., seed=seed, comm=comm)

def generate_survey_data(seed, comm):

    # make the data
    cosmo = cosmology.Planck15
    d = UniformCatalog(nbar=3e-4, BoxSize=512., seed=seed, comm=comm)
    d['RA'], d['DEC'], d['Redshift'] = transform.CartesianToSky(d['Position'], cosmo)

    # make the randoms (ensure nbar is high enough to not have missing values)
    r = UniformCatalog(nbar=3e-4, BoxSize=512., seed=seed*2, comm=comm)
    r['RA'], r['DEC'], r['Redshift'] = transform.CartesianToSky(r['Position'], cosmo)

    return d, r

def reference_sim_tpcf(pos1, redges, BoxSize, randoms=None, pos2=None):
    """Reference 1D 2PCF using halotools"""
    from halotools.mock_observables import tpcf

    estimator = 'Natural' if randoms is None else 'Landy-Szalay'
    do_auto = True if pos2 is None else False
    return tpcf(pos1, redges, period=BoxSize, sample2=pos2, randoms=randoms,
                estimator=estimator, do_auto=do_auto)

def reference_survey_tpcf(data1, randoms1, redges, data2=None, randoms2=None):
    """Reference 1D 2PCF using Corrfunc"""
    from Corrfunc.utils import convert_3d_counts_to_cf
    from Corrfunc.mocks import DDsmu_mocks

    if data2 is None:
        data2 = data1
    if randoms2 is None:
        randoms2 = randoms1

    # updack the columns to pass to Corrfunc
    ra_d1, dec_d1, r_d1 = data1.T
    ra_r1, dec_r1, r_r1 = randoms1.T
    ra_d2, dec_d2, r_d2 = data2.T
    ra_r2, dec_r2, r_r2 = randoms2.T

    # the Corrfunc keywords
    kws = {}
    kws['cosmology'] = 1
    kws['nthreads'] = 1
    kws['nmu_bins'] = 1
    kws['mu_max'] = 1.0
    kws['binfile'] = redges
    kws['is_comoving_dist'] = True

    # do the pair counting
    D1D2 = DDsmu_mocks(0, RA1=ra_d1, DEC1=dec_d1, CZ1=r_d1, RA2=ra_d2, DEC2=dec_d2, CZ2=r_d2, **kws)
    D1R2 = DDsmu_mocks(0, RA1=ra_d1, DEC1=dec_d1, CZ1=r_d1, RA2=ra_r2, DEC2=dec_r2, CZ2=r_r2, **kws)
    D2R1 = DDsmu_mocks(0, RA1=ra_d2, DEC1=dec_d2, CZ1=r_d2, RA2=ra_r1, DEC2=dec_r1, CZ2=r_r1, **kws)
    R1R2 = DDsmu_mocks(0, RA1=ra_r1, DEC1=dec_r1, CZ1=r_r1, RA2=ra_r2, DEC2=dec_r2, CZ2=r_r2, **kws)

    # combine using Landy-Szalay
    ND1 = len(ra_d1); ND2 = len(ra_d2)
    NR1 = len(ra_r1); NR2 = len(ra_r2)

    CF = convert_3d_counts_to_cf(ND1, ND2, NR1, NR2, D1D2, D1R2, D2R1, R1R2)

    return D1D2, D1R2, D2R1, R1R2, CF

@MPITest([4])
def test_sim_periodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)

    # make the bin edges
    redges = numpy.linspace(0.01, 20, 10)

    # compute 2PCF
    r = SimulationBox2PCF('1d', source, redges, periodic=True)

    # verify with halotools
    pos = gather_data(source, "Position")
    cf = reference_sim_tpcf(pos, redges, source.attrs['BoxSize'])
    assert_allclose(cf, r.corr['corr'])


@MPITest([4])
def test_sim_nonperiodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)
    randoms = generate_sim_data(seed=84, comm=comm)

    # make the bin edges
    redges = numpy.linspace(0.01, 20, 10)

    # compute 2PCF
    r = SimulationBox2PCF('1d', source, redges, periodic=False, randoms1=randoms)

    # need randoms if non-periodic!
    with pytest.raises(ValueError):
        r = SimulationBox2PCF('1d', source, redges, periodic=False)

    # verify with reference
    cf = [0.1158045, -0.03074509, -0.0227549, 0.01194295, -0.01553021, -0.001123882, -0.005164478, -0.002296215, 0.003764593]

    # verify with halotools // uses N*N instead of N*(N-1) normalization
    #pos_d = gather_data(source, "Position")
    #pos_r = gather_data(randoms, "Position")
    #cf = reference_sim_tpcf(pos_d, redges, None, randoms=pos_r)
    assert_allclose(cf, r.corr['corr'], rtol=1e-5, atol=1e-5)

@MPITest([4])
def test_sim_periodic_cross(comm):

    # uniform source of particles
    data1 = generate_sim_data(seed=42, comm=comm)
    data2 = generate_sim_data(seed=84, comm=comm)

    # make the bin edges
    redges = numpy.linspace(0.01, 20, 10)

    # compute 2PCF
    r = SimulationBox2PCF('1d', data1, redges, periodic=True, data2=data2)

    # verify with halotools
    pos1 = gather_data(data1, "Position")
    pos2 = gather_data(data2, "Position")
    cf = reference_sim_tpcf(pos1, redges, data1.attrs['BoxSize'], pos2=pos2)
    assert_allclose(cf, r.corr['corr'])

@MPITest([4])
def test_survey_auto(comm):
    cosmo = cosmology.Planck15

    # data and randoms
    data, randoms = generate_survey_data(seed=42, comm=comm)

    # make the bin edges
    redges = numpy.linspace(1.0, 10, 5)

    # compute 2PCF
    r = SurveyData2PCF('1d', data, randoms, redges, cosmo=cosmo)

    # verify with reference
    cf = [0.02063144, -0.03316382, 0.009656524, 0.008670488]
    DD = [1628, 6412, 14774, 26046]
    DR = [1626, 6546, 14562, 25983]
    RR = [1658, 6462, 14492, 26144]

    # run Corrfunc to verify // uses N*N instead of N*(N-1) normalization
    #pos1 = make_corrfunc_input(data, cosmo)
    #pos2 = make_corrfunc_input(randoms, cosmo)
    #DD, DR, _, RR, cf = reference_survey_tpcf(pos1, pos2, redges)
    #DD, DR, RR = DD['npairs'], DR['npairs'], RR['npairs']

    # verify pair counts and CF
    assert_allclose(cf, r.corr['corr'], rtol=1e-7, atol=1e-5)
    assert_allclose(DD, r.D1D2['npairs'])
    assert_allclose(DR, r.D1R2['npairs'])
    assert_allclose(RR, r.R1R2['npairs'])

@MPITest([4])
def test_survey_cross(comm):
    cosmo = cosmology.Planck15

    # data and randoms
    data1, randoms1 = generate_survey_data(seed=42, comm=comm)
    data2, randoms2 = generate_survey_data(seed=84, comm=comm)

    # make the bin edges
    redges = numpy.linspace(0.01, 10.0, 5)

    # compute 2PCF
    r = SurveyData2PCF('1d', data1, randoms1, redges, cosmo=cosmo, data2=data2, randoms2=randoms2)

    # R1R2 computed separately
    R1R2 = SurveyDataPairCount('1d', randoms1, redges, cosmo=cosmo, second=randoms2)
    rGivenR1R2 = SurveyData2PCF('1d', data1, randoms1, redges, cosmo=cosmo, data2=data2, randoms2=randoms2, R1R2=R1R2)
    # verify CF
    assert_allclose(r.corr['corr'], rGivenR1R2.corr['corr'])

    # run Corrfunc to verify
    data1 = make_corrfunc_input(data1, cosmo)
    randoms1 = make_corrfunc_input(randoms1, cosmo)
    data2 = make_corrfunc_input(data2, cosmo)
    randoms2 = make_corrfunc_input(randoms2, cosmo)
    D1D2, D1R2, D2R1, R1R2, cf = reference_survey_tpcf(data1, randoms1, redges,
                                                        data2=data2, randoms2=randoms2)

    # verify pair counts and CF
    assert_allclose(cf, r.corr['corr'])
    assert_allclose(D1D2['npairs'], r.D1D2['npairs'])
    assert_allclose(D1R2['npairs'], r.D1R2['npairs'])
    assert_allclose(D2R1['npairs'], r.D2R1['npairs'])
    assert_allclose(R1R2['npairs'], r.R1R2['npairs'])

@MPITest([1])
def test_low_nbar_randoms(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)
    randoms = UniformCatalog(nbar=3e-6, BoxSize=512., seed=84, comm=comm)

    # make the bin edges
    redges = numpy.linspace(0.01, 5.0, 2)

    # compute 2PCF
    with pytest.warns(UserWarning):
        r = SimulationBox2PCF('1d', source, redges, periodic=False, randoms1=randoms)



import numpy
#from scipy.integrate import odeint
from scipy.integrate import solve_ivp

class LinearNbody:
    """ 
    Perturbations of matter under matter only interaction on an expanding cosmology
    background. Ignoring interaction with radiation.

    This can be considered the linear order perturbative model of hydro nbody-simulations.

    For technical notes, see https://www.overleaf.com/read/pqsgzjssmptb

    Parameters
    ----------
    c_b : float
        baryon velocity in km/s. set to zero to evolve as cdm. Related
        to the thermal history of baryons. 

    background : object
        an object with attributes:
            efunc(z), Omega_b(z), Omega_cdm(z), Omega_ncdm(z) and m_ncdm
        For example, a Cosmology object from nbodykit will work here.

    c_ncdm_1ev_z0 : float
        neutrino velocity in km/s at z=0 for 1ev. The default number 134.423 is widely used.
        set to zero to evolve as cdm.  `c_ncdm = cncdm_1ev_z0 / a / m_ncdm`

    """
    def __init__(self, background, c_b=0, c_ncdm_1ev_z0=134.423):

        self.background = background
        self.c_b = c_b
        self.c_ncdm_1ev_z0 = c_ncdm_1ev_z0

    def efunc(self, z):
        return self.background.efunc(z)

    def Omega_b(self, z):
        return self.background.Omega_b(z)

    def Omega_cdm(self, z):
        return self.background.Omega_cdm(z)

    def Omega_ncdm(self, z):
        return self.background.Omega_ncdm(z)

    @property
    def m_ncdm(self):
        return self.background.m_ncdm

    def integrate(self, k, q0, p0, a, rtol=1e-4):
        """ 
        Solve the 3 fluid model from initial position and momentum q0, and p0, at
        times a.

        This can be slow if neutrino velocity and baryon velocity
        are non-zero. (200,000+ function evaluations with RK45, when rtol=1e-7)

        Parameters
        ----------
        a : array_like
            scale factor requesting the output.

        k : array_like,
             k values to compute the solution, in h/Mpc. From classylss's transfer function.

        q0 : array_like (Nk, 3)
            initial position, -d, produced by `seed` from CLASSylss transfer function
            three species are cdm, baryon and ncdm.
        p0 : array_like (Nk, 3)
            initial momentum, a * v, from `seed`. 
            three species are cdm, baryon and ncdm.
        rtol : float
            relative accuracy. It appears 1e-4 is good for k ~ 10 with reasonable velocities.

        Returns
        -------
        a : array_like (Na), very close but not in general identical to input a.

        q : array_like (Na, Nk, 3)
            position (-d) at different requested a's
        p : array_like (Na, Nk, 3)
            momentum (a * v) at different requested a's. v is peculiar velocity.

        """
        def marshal(q0, p0):
            vector = numpy.concatenate([q0.ravel(), p0.ravel()], axis=0)
            return vector
        # internally work with loga
        lna = numpy.log(a)

        def func(lna, vector):
            a = numpy.exp(lna)
            q = vector[: len(vector) // 2].reshape(q0.shape)
            p = vector[len(vector) // 2:].reshape(p0.shape)
            z = 1. / a - 1.
            dlna = 1.

            E = (self.efunc(z))
            dt = dlna / E
            dp = - numpy.einsum('kij,kj->ki', self.J(k, a), q) * dt

            dq = p / a ** 2 * dt
            #print(a, dp[0] / q[0], p[0] / q[0])
            return marshal(dq, dp)

        v0 = marshal(q0, p0)
        r = solve_ivp(func, (lna[0], lna[-1]), v0,
                           method='RK45', t_eval=lna,
                           rtol=rtol, atol=0, vectorized=False)
        q = r.y[: len(v0) // 2, ...].reshape(list(q0.shape) + [-1])
        p = r.y[len(v0) // 2:, ...].reshape(list(p0.shape) + [-1])
        #print('nfev=', r.nfev)
        return numpy.exp(r.t), q.transpose((2, 0, 1)), p.transpose((2, 0, 1))

    def J(self, k, a):
        """
        The potential term. Use internally in the ODE.
        """
        z = 1./ a - 1.
        E = self.efunc(z)

        Ocdm = self.Omega_cdm(z)
        Ob = self.Omega_b(z)
        Oncdm = self.Omega_ncdm(z)

        if self.c_ncdm_1ev_z0 > 0:
            c_ncdm = (self.c_ncdm_1ev_z0 / self.m_ncdm[0] / a)
        else:
            c_ncdm = 0

        c_b = self.c_b

        j1 = -1.5 * E ** 2 * a ** 2 * numpy.array([
                [Ocdm, Ob, Oncdm,],
                [Ocdm, Ob, Oncdm,],
                [Ocdm, Ob, Oncdm,],
            ])[..., None]

        H0 = 100 # velocity is in km/s, k is in Mpc/h, so H0 is 100.
        j2 = numpy.diag([0, c_b ** 2, c_ncdm **2])[..., None] * k ** 2 / H0 ** 2

        return (j1 + j2).transpose((2, 0, 1))


    @staticmethod
    def seed_from_synchronous(cosmology, a0, Tk=None):
        """
        Convert synchronuous gauge velocity (h_prime) to the momentum in n-body gauge.

        This function repacks the columns to cdm, baryon and ncdm for both q and p, such
        that at a = a0,

        .. code:: python

            q = - d

            p = a v = - a dd / dt

        Parameters
        ----------
        cosmology : object, Cosmology.
            the cosmology object to obtain hubble (with dimension of time unit)
        a0 : float
            the scaling factor to seed, 0.01 for z=99
        Tk : structured array
            use a precomputed transfer function, must be
            the same format as the output of Cosmology.get_transfer(),
            with 'k' in h/Mpc units, 'd_cdm', 'd_b', 'd_ncdm[0]' and
            'h_prime'.

        Returns
        -------
        k: array of (Nk)
        q0: array of (Nk, 3)
        p0: array of (Nk, 3)

        """
        if Tk is None:
            if cosmology.gauge != 'synchronuous':
                cosmology = cosmology.clone(gauge='synchronous')
            Tk = cosmology.get_transfer(1 / a0 - 1)

        # this requires the cosmology object has the same unit of hubble_function
        # as that of potential

        # use CLASS's units to obtain dd/dt.
        H0 = cosmology.hubble_function(0)

        q0 = -numpy.vstack([Tk['d_cdm'], Tk['d_b'], Tk['d_ncdm[0]']]).T
        p0 = numpy.vstack([  0.5 * Tk['h_prime'] * a0 / H0 ,
                      Tk['t_b'] +  0.5 * Tk['h_prime'] * a0 / H0,
                      Tk['t_ncdm[0]'] + 0.5 * Tk['h_prime'] * a0 / H0,
                  ]).T

        return Tk['k'], q0, p0




from nbodykit.source.catalog.uniform import UniformCatalog
from runtests.mpi import MPITest
from nbodykit.lab import CurrentMPIComm
from nbodykit.utils import GatherArray
from numpy.testing import assert_array_equal
from mpi4py import MPI

@MPITest([4])
def test_uniform_invariant(comm):
    cat = UniformCatalog(nbar=2, BoxSize=100., seed=1234, dtype='f4', comm=comm)

    cat1 = UniformCatalog(nbar=2, BoxSize=100., seed=1234, dtype='f4', comm=MPI.COMM_SELF)

    allpos = GatherArray(cat['Position'].compute(), root=Ellipsis, comm=comm)

    assert_array_equal(allpos, cat1['Position'])

    allvel = GatherArray(cat['Velocity'].compute(), root=Ellipsis, comm=comm)

    assert_array_equal(allvel, cat1['Velocity'])



from .base import FileType
from . import tools
from six import string_types
import numpy
try: import fitsio
except ImportError: fitsio = None

class FITSFile(FileType):
    """
    A file object to handle the reading of FITS data using the
    :mod:`fitsio` package.

    See also: https://github.com/esheldon/fitsio

    Parameters
    ----------
    path : str
        the file path to load
    ext: number or string, optional
        The extension.  Either the numerical extension from zero
        or a string extension name. If not sent, data is read from
        the first HDU that has data.
    """
    def __init__(self, path, ext=None):

        # hide the import exception
        if fitsio is None:
            raise ImportError("please install fitsio: ``conda install -c bccp fitsio``")

        self.path = path
        self.dataset = str(ext)

        # try to find the first Table HDU to read if not specified
        with fitsio.FITS(path) as ff:
            if ext is None:
                for i, hdu in enumerate(ff):
                    if hdu.has_data():
                        ext = i
                        break
                if ext is None:
                    raise ValueError("input fits file '%s' has not binary table to read" %path)
            else:
                if isinstance(ext, string_types):
                    if ext not in ff:
                        raise ValueError("FITS file does not contain extension with name '%s'" %ext)
                elif ext >= len(ff):
                    raise ValueError("FITS extension %d is not valid" %ext)

            # make sure we crash if data is wrong or missing
            if not ff[ext].has_data() or ff[ext].get_exttype() == 'IMAGE_HDU':
                raise ValueError("FITS extension %d is not a readable binary table" %ext)

        self.attrs = {}
        self.attrs['ext'] = ext

        # size and dtype
        with fitsio.FITS(path) as ff:
            self.size = ff[ext].get_nrows()
            self.dtype = ff[ext].get_rec_dtype()[0]

    def read(self, columns, start, stop, step=1):
        """
        Read the specified column(s) over the given range

        'start' and 'stop' should be between 0 and :attr:`size`,
        which is the total size of the file

        Parameters
        ----------
        columns : str, list of str
            the name of the column(s) to return
        start : int
            the row integer to start reading at
        stop : int
            the row integer to stop reading at
        step : int, optional
            the step size to use when reading; default is 1

        Returns
        -------
        numpy.array
            structured array holding the requested columns over
            the specified range of rows
        """
        if isinstance(columns, string_types): columns = [columns]

        kws = {'ext':self.attrs['ext'], 'columns':columns, 'rows':range(start, stop, step)}
        return fitsio.read(self.path, **kws)



from nbodykit import set_options, GlobalCache, use_distributed, use_mpi
import pytest
from runtests.mpi import MPITest

def test_bad_options():
    with pytest.raises(KeyError):
        set_options(no_this_option=3)


def test_cache_size():
    with set_options(global_cache_size=100):
        cache = GlobalCache.get()
        assert cache.cache.available_bytes == 100

@MPITest([1, 4])
def test_use_mpi(comm):
    use_mpi(comm)



from runtests.mpi import MPITest
from nbodykit.tutorials import DemoHaloCatalog
from nbodykit import setup_logging
import pytest

setup_logging()

@MPITest([4])
def test_download(comm):

    from halotools.sim_manager import UserSuppliedHaloCatalog

    # download and load the cached catalog
    cat = DemoHaloCatalog('bolshoi', 'rockstar', 0.5, comm=comm)
    assert all(col in cat for col in ['Position', 'Velocity'])

    # convert to halotools catalog
    halotools_cat = cat.to_halotools()
    assert isinstance(halotools_cat, UserSuppliedHaloCatalog)

    # bad simulation name
    with pytest.raises(Exception):
        cat = DemoHaloCatalog('BAD', 'rockstar', 0.5)


@MPITest([4])
def test_download_failure(comm):
    # initialize with bad redshift
    BAD_REDSHIFT = 100.0
    with pytest.raises(Exception):
        cat = DemoHaloCatalog('bolshoi', 'rockstar', BAD_REDSHIFT, comm=comm)



from runtests.mpi import MPITest
from nbodykit.io.tpm import TPMBinaryFile
from nbodykit.io.stack import FileStack
import numpy
import tempfile
import os
import shutil
import contextlib
import pytest

@contextlib.contextmanager
def TemporaryDirectory():
    try:
        tmpdir = tempfile.mkdtemp()
        yield tmpdir
    finally:
        shutil.rmtree(tmpdir)


@MPITest([1])
def test_data(comm):

    with TemporaryDirectory() as tmpdir:
        
        # generate TPM-format data
        pos = numpy.random.random(size=(2048, 3)).astype('f4')
        vel = numpy.random.random(size=(2048, 3)).astype('f4')
        uid = numpy.arange(2048, dtype='u8')
        hdr = numpy.ones(28, dtype='?')
        
        for i, name in enumerate(['tpm.000', 'tpm.001']):
            sl = slice(i*1024, (i+1)*1024)
            
            # write to file
            fname = os.path.join(tmpdir, name)
            with open(fname, 'wb') as ff:
                hdr.tofile(ff)
                pos[sl].tofile(ff); vel[sl].tofile(ff); uid[sl].tofile(ff)
            
        # initialize the stack
        path = os.path.join(tmpdir, 'tpm.00*')
        f = FileStack(TPMBinaryFile, path, precision='f4')
        
        # check size
        assert f.size == 2048
        
        # and data
        numpy.testing.assert_almost_equal(pos, f['Position'][:])
        numpy.testing.assert_almost_equal(vel, f['Velocity'][:])
        numpy.testing.assert_almost_equal(uid, f['ID'][:])
        
        # pass a list
        paths = [os.path.join(tmpdir, f) for f in ['tpm.000', 'tpm.001']]
        f = FileStack(TPMBinaryFile, paths, precision='f4')
        
        # check size
        assert f.size == 2048
        assert f.nfiles == 2
        
        # and add and attrs
        f.attrs['size'] = 2048
        

@MPITest([1])
def test_single_path(comm):

    with TemporaryDirectory() as tmpdir:
        
        # generate TPM-format data
        pos = numpy.random.random(size=(2048, 3)).astype('f4')
        vel = numpy.random.random(size=(2048, 3)).astype('f4')
        uid = numpy.arange(2048, dtype='u8')
        hdr = numpy.ones(28, dtype='?')
        
        for i, name in enumerate(['tpm.000', 'tpm.001']):
            sl = slice(i*1024, (i+1)*1024)
            
            # write to file
            fname = os.path.join(tmpdir, name)
            with open(fname, 'wb') as ff:
                hdr.tofile(ff)
                pos[sl].tofile(ff); vel[sl].tofile(ff); uid[sl].tofile(ff)

        # single path
        f = FileStack(TPMBinaryFile, os.path.join(tmpdir, 'tpm.000'), precision='f4')
        assert f.size == 1024
        assert f.nfiles == 1


@MPITest([1])
def test_bad_path(comm):

    with TemporaryDirectory() as tmpdir:
        
        # generate TPM-format data
        pos = numpy.random.random(size=(2048, 3)).astype('f4')
        vel = numpy.random.random(size=(2048, 3)).astype('f4')
        uid = numpy.arange(2048, dtype='u8')
        hdr = numpy.ones(28, dtype='?')
        
        for i, name in enumerate(['tpm.000', 'tpm.001']):
            sl = slice(i*1024, (i+1)*1024)
            
            # write to file
            fname = os.path.join(tmpdir, name)
            with open(fname, 'wb') as ff:
                hdr.tofile(ff)
                pos[sl].tofile(ff); vel[sl].tofile(ff); uid[sl].tofile(ff)
            
        # bad path name
        with pytest.raises(ValueError): 
            f = FileStack(TPMBinaryFile, ff, precision='f4')



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import kdcount.correlate as correlate
import os
import pytest

setup_logging()

def gather_data(source, name):
    return numpy.concatenate(source.comm.allgather(source[name].compute()), axis=0)

def generate_sim_data(seed, comm):
    return UniformCatalog(nbar=3e-6, BoxSize=512., seed=seed, comm=comm)

def generate_survey_data(seed, comm):
    cosmo = cosmology.Planck15
    s = RandomCatalog(1000, seed=seed, comm=comm)

    # ra, dec, z
    s['Redshift'] = s.rng.normal(loc=0.5, scale=0.1)
    s['RA'] = s.rng.uniform(low=110, high=260)
    s['DEC'] = s.rng.uniform(low=-3.6, high=60.)

    # position
    s['Position'] = transform.SkyToCartesian(s['RA'], s['DEC'], s['Redshift'], cosmo=cosmo)

    return s

def reference_sim_paircount(pos1, w1, redges, Nmu, boxsize, pos2=None, w2=None, los=2):
    """Reference pair counting via kdcount"""

    tree1 = correlate.points(pos1, boxsize=boxsize, weights=w1)
    if pos2 is None:
        tree2 = tree1
    else:
        tree2 = correlate.points(pos2, boxsize=boxsize, weights=w2)

    bins = correlate.FlatSkyBinning(redges, Nmu, los=los, mu_min=0., absmu=True,)
    pc = correlate.paircount(tree1, tree2, bins, np=0, compute_mean_coords=True)
    return numpy.nan_to_num(pc.pair_counts), numpy.nan_to_num(pc.mean_centers[0]), pc.sum1

def reference_survey_paircount(pos1, w1, redges, Nmu, pos2=None, w2=None, los=2):
    """Reference pair counting via kdcount"""

    tree1 = correlate.points(pos1, boxsize=None, weights=w1)
    if pos2 is None:
        tree2 = tree1
    else:
        tree2 = correlate.points(pos2, boxsize=None, weights=w2)

    bins = correlate.RmuBinning(redges, Nmu, observer=(0,0,0), mu_min=0., absmu=True)
    pc = correlate.paircount(tree1, tree2, bins, np=0, compute_mean_coords=True)
    return numpy.nan_to_num(pc.pair_counts), numpy.nan_to_num(pc.mean_centers[0]), pc.sum1


@MPITest([1, 3])
def test_sim_periodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    Nmu = 10
    r = SimulationBoxPairCount('2d', source, redges, periodic=True, weight='Weight', Nmu=Nmu)

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = reference_sim_paircount(pos, w, redges, Nmu, source.attrs['BoxSize'])
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])

@MPITest([3])
def test_sim_diff_los(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    Nmu = 2
    r = SimulationBoxPairCount('2d', source, redges, periodic=True, weight='Weight', Nmu=Nmu, los='x')

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = reference_sim_paircount(pos, w, redges, Nmu, source.attrs['BoxSize'], los=0)
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])

@MPITest([1, 3])
def test_sim_nonperiodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)
    source['Weight'] = numpy.random.random(size=len(source))

    # make the bin edges
    redges = numpy.linspace(10, 40., 10)

    # do the weighted paircount
    Nmu = 50
    r = SimulationBoxPairCount('2d', source, redges, periodic=False, weight='Weight', Nmu=Nmu)

    pos = gather_data(source, "Position")
    w = gather_data(source, "Weight")

    # verify with kdcount
    npairs, ravg, wsum = reference_sim_paircount(pos, w, redges, Nmu, None)
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])


@MPITest([1, 3])
def test_sim_periodic_cross(comm):

    # generate data
    first = generate_sim_data(seed=42, comm=comm)
    second = generate_sim_data(seed=84, comm=comm)

    # make the bin edges
    redges = numpy.linspace(10, 40., 10)

    # do the paircount
    Nmu = 10
    r = SimulationBoxPairCount('2d', first, redges, second=second, periodic=True, Nmu=Nmu)

    pos1 = gather_data(first, "Position")
    pos2 = gather_data(second, "Position")

    # verify with kdcount
    npairs, ravg, wsum = reference_sim_paircount(pos1, None, redges, Nmu, first.attrs['BoxSize'], pos2=pos2)
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])

@MPITest([1, 4])
def test_survey_auto(comm):
    cosmo = cosmology.Planck15

    # random particles
    source = generate_survey_data(seed=42, comm=comm)
    source['Weight'] = source.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 1000., 10)

    # do the weighted paircount
    Nmu = 50
    r = SurveyDataPairCount('2d', source, redges, cosmo, weight='Weight', Nmu=Nmu)

    pos = gather_data(source, 'Position')
    w = gather_data(source, 'Weight')

    # verify with kdcount
    npairs, ravg, wsum = reference_survey_paircount(pos, w, redges, Nmu)
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])


@MPITest([1, 4])
def test_survey_cross(comm):
    cosmo = cosmology.Planck15

    # random particles
    first = generate_survey_data(seed=42, comm=comm)
    first['Weight'] = first.rng.uniform()
    second = generate_survey_data(seed=84, comm=comm)
    second['Weight'] = second.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the paircount
    Nmu = 5
    r = SurveyDataPairCount('2d', first, redges, cosmo, second=second, Nmu=Nmu)

    pos1 = gather_data(first, 'Position')
    pos2 = gather_data(second, 'Position')
    w1 = gather_data(first, 'Weight')
    w2 = gather_data(second, 'Weight')

    # verify with kdcount
    npairs, ravg, wsum = reference_survey_paircount(pos1, w1, redges, Nmu, pos2=pos2, w2=w2)
    assert_allclose(ravg, r.pairs['r'])
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['wnpairs'])

    # test save
    r.save('paircount-test.json')
    r2 = SurveyDataPairCount.load('paircount-test.json', comm=comm)
    assert_array_equal(r.pairs.data, r2.pairs.data)

    if comm.rank == 0: os.remove('paircount-test.json')

@MPITest([1])
def test_missing_Nmu(comm):

    # generate data
    source = generate_sim_data(seed=42, comm=comm)
    redges = numpy.linspace(10, 150, 10)

    # missing Nmu
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('2d', source, redges)

    # wrong mode
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, redges, Nmu=10)



from nbodykit.base.catalog import CatalogSource
from pmesh.domain import GridND
from nbodykit.utils import split_size_3d
import numpy

class SubVolumesCatalog(CatalogSource):
    """ A catalog that distributes the particles spatially into subvolumes per
        MPI rank.

        Attributes
        ----------
        domain : :class:`pmesh.domain.GridND`;
            The domain objects for decomposition. If None, generate
            a domain to decompose the catalog into a 3d grid.

        layout : A large object that holds which particle belongs to which rank.
        source : the original source object

        Parameters
        ----------
        columns: list
            a list of columns to already exchange
    """
    def __init__(self, source, domain=None, position='Position', columns=None):
        comm = source.comm

        if domain is None:
            # determine processor division for domain decomposition
            np = split_size_3d(comm.size)

            if comm.rank == 0:
                self.logger.info("using cpu grid decomposition: %s" %str(np))

            grid = [
                numpy.linspace(0, source.attrs['BoxSize'][0], np[0] + 1, endpoint=True),
                numpy.linspace(0, source.attrs['BoxSize'][1], np[1] + 1, endpoint=True),
                numpy.linspace(0, source.attrs['BoxSize'][2], np[2] + 1, endpoint=True),
            ]

            domain = GridND(grid, comm=comm)

        self.domain = domain
        self.source = source

        layout = domain.decompose(source[position].compute())

        self._size = layout.newlength

        CatalogSource.__init__(self, comm=comm)
        self.attrs.update(source.attrs)

        self._frozen = {}
        if columns is None: columns = source.columns

        for column in columns:
            data = source[column].compute()
            self._frozen[column] = self.make_column(layout.exchange(data))

    @property
    def hardcolumns(self):
        return sorted(list(self._frozen.keys()))

    def get_hardcolumn(self, col):
        return self._frozen[col]



from .base import FileType
from . import tools
from six import string_types
import numpy
import os
from collections import  namedtuple

try: import h5py
except ImportError: h5py = None

ColumnInfo = namedtuple('ColumnInfo', ['size', 'dtype', 'dset'])

def find_datasets(info, attrs, name, obj):
    """
    Recursively add a ``ColumnInfo`` named tuple to the ``info`` dict
    if ``obj`` is a Dataset

    When ``obj`` is a structured array with named fields, a
    ``ColumnInfo`` tuple will be added for each of the named fields
    """
    # only gather info on dataset
    if isinstance(obj, h5py.Dataset):

        # update meta-data (remember: all strings in h5py stored encoded data)
        attrs[str(name)] = {str(k):obj.attrs[k] for k in obj.attrs}

        # structured array
        if obj.dtype.kind == 'V':
            for col in obj.dtype.names:
                size = len(obj)
                dtype = obj.dtype[col]
                key = str(os.path.join(name, col))
                info[key] = ColumnInfo(size=size, dtype=dtype, dset=name)
        # normal array
        else:
            size = obj.shape[0]
            subshape = obj.shape[1:]
            fmt = obj.dtype.type
            if len(subshape): fmt = (fmt,) + subshape
            dtype = numpy.dtype(fmt)
            key = str(name)
            info[key] = ColumnInfo(size=size, dtype=dtype, dset=name)

class HDFFile(FileType):
    """
    A file object to handle the reading of columns of data from a :mod:`h5py`
    HDF5 file.

    See http://docs.h5py.org for documentation on :mod:`h5py`.

    Parameters
    ----------
    path : str
        the file path to load
    root : str, optional
        the start path in the HDF file, loading all data below this path
    exclude : list of str, optional
        list of path names to exclude; these can be absolute paths, or paths
        relative to ``root``
    """
    def __init__(self, path, dataset='/', exclude=[], header=None, root=None):

        if h5py is None:
            raise ImportError("please install h5py to use HDFFile")

        self.path = path

        if root is not None:
            import warnings
            warnings.warn("Use dataset= argument, not root= ", DeprecationWarning, 2)
            dataset = root

        self.dataset = dataset
        self.attrs = {}

        # gather dtype and size information from file
        info = {}
        with h5py.File(self.path, 'r') as ff:

            # make sure root and any excluded paths are valid
            if self.dataset not in ff:
                raise ValueError("'%s' is not a valid path in HDF file" % self.dataset)

            # verify and format the excluded names
            _exclude = []
            for excluded in exclude:
                if excluded not in ff:
                    if os.path.join(self.dataset, excluded) not in ff:
                        raise ValueError("'%s' is not a valid path name; cannot be excluded" %excluded)
                    else:
                        excluded = os.path.join(self.dataset, excluded)
                _exclude.append(excluded.lstrip('/'))

            if header is not None:
                if header not in ff:
                    raise ValueError("'%s' is not a valid path in HDF file" % header)
                ds = ff[header]
                for key in ds.attrs:
                    self.attrs[key] = ds.attrs[key]
                _exclude.append(header)

            # get the info about possible columns
            sub = ff[self.dataset]
            if isinstance(sub, h5py.Dataset):
                find_datasets(info, self.attrs, '', sub)
            else:
                sub.visititems(lambda *args: find_datasets(info, self.attrs, *args))

        # exclude columns
        for col in list(info):
            absname = os.path.join(self.dataset, col)
            if any(absname.lstrip('/').startswith(ex) for ex in _exclude):
                self.logger.info("ignoring excluded column '%s'" %col)
                info.pop(col)

        # verify all the datasets have a single size
        sizes = set([info[col].size for col in info])
        if len(sizes) > 1:
            msg = "size mismatch in datasets of file; please use ``exclude`` to remove datasets of the wrong size\n"
            msg += "\n".join(["size of '%s': %d" %(col, info[col].size) for col in info])
            raise ValueError(msg)

        # empty file check
        if not len(sizes):
            raise ValueError("HDF file appears to contain datasets")
        self.size = list(sizes)[0]

        # if single Dataset with structured array, allow relative names
        unique_dsets = set([info[col].dset for col in info])
        single_structured_arr = len(unique_dsets) == 1 and len(info) > 1

        # construct the data type from "info"
        dtype = []
        for col in info:
            name = col
            if single_structured_arr:
                name = name.rsplit('/', 1)[-1]
            dtype.append((name, info[col].dtype))
        self.dtype = numpy.dtype(dtype)

        # set the root properly if columns stored as single structured array
        if single_structured_arr:
            name = list(unique_dsets)[0]
            self.dataset = os.path.join(self.dataset, name)
            self.attrs = self.attrs[name]
            self.logger.info("detected single structured array stored as dataset; changing root of HDF file to %s" %self.dataset)

    def read(self, columns, start, stop, step=1):
        """
        Read the specified column(s) over the given range

        'start' and 'stop' should be between 0 and :attr:`size`,
        which is the total size of the file

        Parameters
        ----------
        columns : str, list of str
            the name of the column(s) to return
        start : int
            the row integer to start reading at
        stop : int
            the row integer to stop reading at
        step : int, optional
            the step size to use when reading; default is 1

        Returns
        -------
        numpy.array
            structured array holding the requested columns over
            the specified range of rows
        """
        if isinstance(columns, string_types): columns = [columns]

        dt = [(col, self.dtype[col]) for col in columns]
        toret = numpy.empty(tools.get_slice_size(start, stop, step), dtype=dt)

        with h5py.File(self.path, 'r') as ff:
            # compile a list of datasets
            dsets = {}

            for col in columns:

                # absolute name of column (with root path prepended)
                name = os.path.join(self.dataset, col)

                if name in ff:
                    # data from a h5py Dataset directly
                    dsets[name] = [(col, None)]
                    continue
                else:
                    # data from a column in a structured array
                    splitcol = name.rsplit('/', 1)
                    if len(splitcol) != 2:
                        raise ValueError("error trying to access column '%s' in HDF file" %col)

                    name, field = splitcol

                    try:
                        dsets[name].append((col, field))
                    except KeyError:
                        dsets[name] = [(col, field)]
            # then read through the list of datasets,
            # columns in the same dataset is read only once.
            # see, http://docs.h5py.org/en/latest/high/dataset.html#reading-writing-data

            # it is a bit ugly seems to work will see if this fixes the slowness.
            for name, cols in dsets.items():
                dset = ff[name]
                if len(cols) == 1 and cols[0][1] is None:
                    [[col, field]] = cols
                    toret[col][:] = dset[start:stop:step]
                else:
                    fields = [field for col, field in cols]
                    fields.append(slice(start, stop, step))
                    results = dset[tuple(fields)]

                    for col, field in cols:
                        if len(cols) > 1:
                            toret[col][:] = results[field]
                        else:
                            toret[col][:] = results

        return toret



from classylss.binding import ClassEngine, Background, Spectra, Perturbs, Primordial, Thermo
from classylss.astropy_compat import AstropyCompat

import numpy
from six import string_types
import os
import functools

def store_user_kwargs():
    """
    Decorator that adds the ``_user_kwargs`` attribute to the class to track
    which arguments the user actually supplied.
    """
    def decorator(function):
        @functools.wraps(function)
        def inner(self, *args, **kwargs):
            self._user_args = args
            self._user_kwargs = kwargs
            return function(self, *args, **kwargs)
        return inner
    return decorator


class Cosmology(object):
    r"""
    A cosmology calculator based on the CLASS binding in :mod:`classylss`.

    It is a collection of all method provided by the CLASS interfaces.
    The object is immutable. To obtain an instance with a new set of parameters
    use :func:`clone` or :func:`match`.

    The individual interfaces can be accessed too, such that
    `c.Spectra.get_transfer` and `c.get_transfer` are identical.

    .. important::

        A default set of units is assumed. Those units are:

        * temperature: :math:`\mathrm{K}`
        * distance: :math:`h^{-1} \mathrm{Mpc}`
        * wavenumber: :math:`h \mathrm{Mpc}^{-1}`
        * power: :math:`h^{-3} \mathrm{Mpc}^3`
        * density: :math:`10^{10} (M_\odot/h) (\mathrm{Mpc}/h)^{-3}`
        * neutrino mass: :math:`\mathrm{eV}`
        * time: :math:`\mathrm{Gyr}`
        * :math:`H_0`: :math:`(\mathrm{km} \ \mathrm{s^{-1}}) / (h^{-1} \ \mathrm{Mpc})`

    Notes
    -----
    * The default configuration assumes a flat cosmology, :math:`\Omega_{0,k}=0`.
      Pass ``Omega0_k`` as a keyword to specify the desired non-flat curvature.
    * For consistency of variable names, the present day values can be passed
      with or without '0' postfix, e.g., ``Omega0_cdm`` is translated to
      ``Omega_cdm`` as CLASS always uses the names without `0` as input
      parameters.
    * By default, a cosmological constant (``Omega0_lambda``) is assumed, with
      its density value inferred by the curvature condition.
    * Non-cosmological constant dark energy can be used by specifying the
      ``w0_fld``, ``wa_fld``, and/or ``Omega_fld`` values.
    * To pass in CLASS parameters that are not valid Python argument names, use
      the dictionary/keyword arguments trick, e.g.
      ``Cosmology(..., **{'temperature contributions': 'y'})``
    * ``Cosmology(**dict(c))`` is not supposed to work; use ``Cosmology.from_dict(dict(c))``.

    Parameters
    ----------
    h : float
        the dimensionless Hubble parameter
    T0_cmb : float
        the temperature of the CMB in Kelvins
    Omega0_b : float
        the current baryon density parameter, :math:`\Omega_{b,0}`. Currently
        unrealistic cosmology where Omega_b == 0 is not supported.
    Omega0_cdm : float
        the current cold dark matter density parameter, :math:`\Omega_{cdm,0}`
    N_ur : float
        the number of ultra-relativistic (massless neutrino) species; the
        default number is inferred based on the number of massive neutrinos
        via the following logic: if you have respectively 1,2,3 massive
        neutrinos and use the default ``T_ncdm`` value (0.71611 K), designed
        to give m/omega of 93.14 eV, and you wish to have ``N_eff=3.046`` in
        the early universe, then ``N_ur`` is set to 2.0328, 1.0196, 0.00641,
        respectively.
    m_ncdm : list, None
        the masses (in eV) for all massive neutrino species; an empty list
        should  be passed for no massive neutrinos. The default is a single
        massive neutrino with mass of 0.06 eV
    P_k_max : float
        the maximum ``k`` value to compute power spectrum results to, in units
        of :math:`h/Mpc`
    P_z_max : float
        the maximum redshift to compute power spectrum results to
    gauge : str,
        either synchronous or newtonian
    n_s : float
        the tilt of the primordial power spectrum
    nonlinear : bool
        whether to compute nonlinear power spectrum results via HaloFit
    verbose : bool
        whether to turn on the default CLASS logging for all submodules
    **kwargs :
        extra keyword parameters to pass to CLASS. Mainly used to pass-in
        parameter names that are not valid Python function argument names,
        e.g. ``temperature contributions``, or ``number count contributions``.
        Users should be wary of configuration options that may conflict
        with the base set of parameters. To override parameters, chain the
        result with :func:`clone`.
    """
    # delegate resolve order -- a pun at mro; which in
    # this case introduces the meta class bloat and doesn't solve
    # the issue. We want delayed initialization of interfaces
    # or so-called 'mixins'.
    # easier to just use delegates with a customized getattr.
    # this doesn't work well with automated documentation tools though,
    # unfortunately.

    dro = [AstropyCompat, Thermo, Spectra, Perturbs, Primordial, Background, ClassEngine]
    dro_dict = dict([(n.__name__, n) for n in dro])

    @store_user_kwargs()
    def __init__(self,
            h=0.67556,
            T0_cmb=2.7255,
            Omega0_b=0.022032/0.67556**2,
            Omega0_cdm=0.12038/0.67556**2,
            N_ur=None,
            m_ncdm=[0.06],
            P_k_max=10.,
            P_z_max=100.,
            gauge='synchronous',
            n_s=0.9667,
            nonlinear=False,
            verbose=False,
            **kwargs # additional arguments to pass to CLASS
        ):

        # quickly copy over all arguments --
        # at this point locals only contains the arguments.
        args = dict(locals())

        # store the extra CLASS params
        kwargs = args.pop('kwargs')

        # remove some non-CLASS variables
        args.pop('self')

        # check for deprecated init signature
        deprecated_args = check_deprecated_init(self._user_args, self._user_kwargs)
        if deprecated_args is not None:

            # check for conflicts between named args user passed and
            # deprecated args passed via **kwargs
            for a in deprecated_args:
                if a in self._user_kwargs:
                    raise ValueError("Parameter conflicts; use '%s' parameter only" %a)

            # if we make it here, it is a valid deprecated syntax
            import warnings
            warnings.warn(("This init signature is deprecated; see the Cosmology "
                           "docstring for new signature"), FutureWarning)
            args = deprecated_args
        else:
            # merge the kwargs; without resolving conflicts.
            args.update(kwargs)

        # check for input conflicts (using kwargs user actually input)
        check_args(self._user_kwargs)

        # verify and set defaults
        pars = compile_args(args)

        # use set state to de-serialize the object.
        self.__setstate__(pars)

    def __str__(self):
        """
        Return a dict string when printed
        """
        return dict(self).__str__()
    
    def __iter__(self):
        """
        Allows dict() to be used on class.
        Use :func:`from_dict` to reconstruct an instance.
        """
        pars = self.pars.copy()
        for k in pars:
            yield k, pars[k]

    def __dir__(self):
        """ a list of all members from all delegate classes """
        r = []
        # first allow tab completion of delegate names; to help resolve conflicts
        r.extend([n.__name__ for n in self.dro])
        # then allow tab completion of all delegate methods
        for i in reversed(self.dro):
            r.extend(dir(i))
        return sorted(list(set(r)))

    def __setattr__(self, key, value):

        # do not allow setting of properties of the delegate classes
        if any(hasattr(n, key) for n in self.dro):
            raise ValueError(("the Cosmology object is immutable; use clone() or "
                              "match() to update parameters"))

        return object.__setattr__(self, key, value)

    def __getattr__(self, name):
        """
        Find the proper delegate, initialize it, and run the method
        """
        # getting a delegate explicitly, e.g. c.Background
        if name in self.dro_dict:
            iface = self.dro_dict[name]
            if iface not in self.delegates:
                self.delegates[iface] = iface(self.engine)
            return self.delegates[iface]

        # resolving a name from the delegates : c.Om0 => c.Background.Om0
        for iface in self.dro:
            if hasattr(iface, name):
                if iface not in self.delegates:
                    self.delegates[iface] = iface(self.engine)
                d = self.delegates[iface]
                return getattr(d, name)
        else:
            raise AttributeError("Attribute `%s` not found in any of the delegate objects" % name)

    def __getstate__(self):
        return (self.pars)

    @property
    def sigma8(self):
        """
        The amplitude of matter fluctuations at :math:`z=0` in a sphere
        of radius :math:`r = 8 \ h^{-1}\mathrm{Mpc}`.

        This is not an input CLASS parameter. To scale ``sigma8``, use
        :func:`match`, which adjusts scalar amplitude ``A_s`` to
        achieve the desired ``sigma8``.
        """
        return self.Spectra.sigma8

    @property
    def Omega0_cb(self):
        """
        The total density of CDM and Baryon.

        This is not an input CLASS parameter. To scale ``Omega0_cb``, use
        :func:`match`.
        """
        return self.Background.Omega0_cdm + self.Background.Omega0_b

    def match(self, sigma8=None, Omega0_cb=None, Omega0_m=None):
        """
        Creates a new cosmology that matches a derived parameter. This is different
        from clone, where CLASS parameters are used.

        Note that we only supoort matching one derived parameter at a time,
        because the matching is in general non-commutable.

        Parameters
        ----------
        sigma8 : float or None
            We scale the scalar amplitude ``A_s`` to achieve the desired ``sigma8``.
        Omega0_cb: float or None
            Desired total energy density of CDM and baryon.
        Omega0_m: float or None
            Desired total energy density of matter-like components (included ncdm)

        Returns
        -------
        A new cosmology parameter where the derived parameter matches the given constrain.

        """

        if sum(0 if i is None else 1 for i in [sigma8, Omega0_cb, Omega0_m]) != 1:
            raise ValueError("Only match one derived parameter at one time; but multiple is given.")

        if sigma8 is not None:
            return self.clone(A_s=self.A_s * (sigma8/self.sigma8)**2)

        if Omega0_cb is not None:
            rat = Omega0_cb / self.Omega0_cb
            return self.clone(Omega_b=rat * self.Omega0_b, Omega_cdm=rat * self.Omega0_cdm)

        if Omega0_m is not None:
            Omega0_cb = Omega0_m - (self.Omega0_ncdm_tot - self.Omega0_pncdm_tot) - self.Omega0_dcdm
            return self.match(Omega0_cb=Omega0_cb)

        return self

    def to_astropy(self):
        """
        Initialize and return a subclass of :class:`astropy.cosmology.FLRW`
        from the :class:`Cosmology` class.

        Returns
        -------
        subclass of :class:`astropy.cosmology.FLRW` :
            the astropy class holding the cosmology values
        """
        import astropy.cosmology as ac
        import astropy.units as au

        is_flat = True
        needs_w0 = False
        needs_wa = False

        pars = {}
        pars['H0'] = 100*self.h
        pars['Om0'] = self.Omega0_b + self.Omega0_cdm # exclude massive neutrinos to better match astropy
        pars['Tcmb0'] = self.Tcmb0
        pars['Neff'] = self.Neff
        pars['Ob0'] = self.Ob0

        if self.has_massive_nu:

            # all massless by default
            m_nu = numpy.zeros(int(numpy.floor(self.Neff)))

            # then add massive species
            m_nu[:len(self.m_ncdm)] = self.m_ncdm[:]
            pars['m_nu'] = au.Quantity(m_nu, au.eV)

        if self.Ok0 != 0.:
            pars['Ode0'] = self.Ode0
            is_flat = False

        if self.wa_fld != 0:
            pars['wa'] = self.wa_fld
            pars['w0'] = self.wa_fld
            needs_wa = True

        if self.w0_fld != -1:
            pars['w0'] = self.w0_fld
            needs_w0 = True

        # determine class to return
        prefix = "" if not is_flat else "Flat"
        if needs_wa:
            cls = prefix + "w0waCDM"
        elif needs_w0:
            cls = prefix + "wCDM"
        else:
            cls = prefix + "LambdaCDM"
        cls = getattr(ac, cls)

        return cls(**pars)

    @classmethod
    def from_astropy(kls, cosmo, **kwargs):
        """
        Initialize and return a :class:`Cosmology` object from a subclass of
        :class:`astropy.cosmology.FLRW`.

        Parameters
        ----------
        cosmo : subclass of :class:`astropy.cosmology.FLRW`.
            the astropy cosmology instance
        **kwargs :
            extra keyword parameters to pass when initializing;
            they shall not be in conflict with the parameters
            inferred from cosmo. To override parameters,
            chain the result with :func:`clone`.

        Returns
        -------
        :class:`Cosmology` :
            the initialized cosmology object
        """
        args = astropy_to_dict(cosmo)
        # merge in additional arguments -- this will die if
        # there are conflicts.
        args.update(kwargs)

        # astropy_to_dict creates args, so we can use the 'user-friendly'
        # constructor.
        return Cosmology(**args)

    @classmethod
    def from_file(cls, filename, **kwargs):
        """
        Initialize a :class:`Cosmology` object from the CLASS parameter file

        Parameters
        ----------
        filename : str
            the name of the parameter file to read
        **kwargs :
            extra keyword parameters to pass when initializing;
            they shall not be in conflict with the parameters
            inferred from cosmo. To override parameters,
            chain the result with :func:`clone`.
        """
        from classylss import load_ini

        # extract dictionary of parameters from the file
        pars = load_ini(filename)

        # intentionally not using merge; use clone if
        # parameters are to modified.
        pars.update(kwargs)

        return cls.from_dict(pars)

    @classmethod
    def from_dict(kls, pars):
        """
        Creates a Cosmology from a pars dictionary.

        This is a rather 'raw' API.
        The dictionary must be readable by ClassEngine.
        Unlike ``Cosmology(**args)``, ``pars`` must
        not contain any convenient names defined here.
        """
        self = object.__new__(Cosmology)
        self.__setstate__(pars)
        return self

    def __setstate__(self, state):

        pars = state

        # initialize the engine as the backup delegate.
        self.engine = ClassEngine(pars)
        self.delegates = {ClassEngine: self.engine}
        self.pars = pars

    def clone(self, **kwargs):
        """
        Create a new cosmology based on modification of self, with the
        input keyword parameters changed.

        Parameters
        ----------
        **kwargs :
            keyword parameters to adjust

        Returns
        -------
        :class:`Cosmology`
            a copy of self, with the input ``kwargs`` adjusted
        """
        # this call to merge_args is OK because self.pars is
        # a valid set of args
        args = merge_args(self.pars, kwargs)
        check_args(args)
        pars = compile_args(args)

        return type(self).from_dict(pars)

def astropy_to_dict(cosmo):
    """
    Convert an astropy cosmology object to a dictionary of parameters
    suitable for initializing a Cosmology object.
    """
    from astropy import cosmology, units

    args = {}
    args['h'] = cosmo.h
    args['T0_cmb'] = cosmo.Tcmb0.value
    if cosmo.Ob0 is not None:
        args['Omega0_b'] = cosmo.Ob0
    else:
        raise ValueError("please specify a value 'Ob0' ")
    args['Omega0_cdm'] = cosmo.Om0 - cosmo.Ob0 # should be okay for now

    # handle massive neutrinos
    if cosmo.has_massive_nu:

        # convert to eV
        m_nu = cosmo.m_nu
        if hasattr(m_nu, 'unit') and m_nu.unit != units.eV:
            m_nu = m_nu.to(units.eV)
        else:
            m_nu = units.eV * m_nu
        # from CLASS notes:
        # one more remark: if you have respectively 1,2,3 massive neutrinos,
        # if you stick to the default value pm equal to 0.71611, designed to give m/omega of
        # 93.14 eV, and if you want to use N_ur to get N_eff equal to 3.046 in the early universe,
        # then you should pass here respectively 2.0328,1.0196,0.00641
        N_ur = [2.0328, 1.0196, 0.00641]
        N_massive = (m_nu > 0.).sum()
        args['N_ur'] = (cosmo.Neff/3.046) * N_ur[N_massive-1]

        args['m_ncdm'] = [k.value for k in sorted(m_nu[m_nu > 0.], reverse=True)]
    else:
        args['m_ncdm'] = []
        args['N_ur'] = cosmo.Neff

    # specify the curvature
    args['Omega0_k'] = cosmo.Ok0

    # handle dark energy
    if isinstance(cosmo, cosmology.LambdaCDM):
        pass
    elif isinstance(cosmo, cosmology.wCDM):
        args['w0_fld'] = cosmo.w0
        args['wa_fld'] = 0.
        args['Omega0_Lambda'] = 0. # use Omega_fld
    elif isinstance(cosmo, cosmology.w0waCDM):
        args['w0_fld'] = cosmo.w0
        args['wa_fld'] = cosmo.wa
        args['Omega0_Lambda'] = 0. # use Omega_fld
    else:
        cls = cosmo.__class__.__name__
        valid = ["LambdaCDM", "wCDM", "w0waCDM"]
        msg = "dark energy equation of state not recognized for class '%s'; " %cls
        msg += "valid classes: %s" %str(valid)
        raise ValueError(msg)

    return args

def compile_args(args):
    """
    Compile the input args of Cosmology object to the input parameters (pars) to
    a :class:`Cosmology` object.

    A variety of defaults are set to tune CLASS for quantities used in
    large scale structures.

    Difference between pars and args:
     - anything that is valid pars is also valid args.
     - after replacing our customizations in args, we get pars.

    Note that CLASS will check for additional conflicts.

    see :func:`merge_args`
    """
    pars = {} # we try to make pars write only.

    # set some default parameters
    pars.setdefault('output', "vTk dTk mPk")
    pars.setdefault('extra metric transfer functions', 'y')

    # args and pars are pretty much compatible;
    pars.update(args)

    def set_alias(pars_name, args_name):
        if args_name not in args: return
        v = args[args_name]
        pars.pop(args_name) # pop because we copied everything.
        if pars_name in args:
            v = args[pars_name]
        pars[pars_name] = v

    set_alias('T_cmb', 'T0_cmb')
    set_alias('Omega_cdm', 'Omega0_cdm')
    set_alias('Omega_b', 'Omega0_b')
    set_alias('Omega_k', 'Omega0_k')
    set_alias('Omega_ur', 'Omega0_ur')
    set_alias('Omega_Lambda', 'Omega_lambda') # classylss variable has lowercase l
    set_alias('Omega_Lambda', 'Omega0_lambda') # classylss variable has lowercase l
    set_alias('Omega_Lambda', 'Omega0_Lambda')
    set_alias('Omega_fld', 'Omega0_fld')
    set_alias('Omega_ncdm', 'Omega0_ncdm')
    set_alias('Omega_g', 'Omega0_g')

    # turn on verbosity
    if 'verbose' in args:
        pars.pop('verbose')
        verbose = args['verbose']
        if verbose:
            for par in ['input', 'background', 'thermodynamics', 'perturbations',
                        'transfer', 'primordial', 'spectra', 'nonlinear', 'lensing']:
                name = par + '_verbose'
                if name not in pars: pars[name] = 1

    # no massive neutrinos
    if 'm_ncdm' in args:
        pars.pop('m_ncdm')
        m_ncdm = args['m_ncdm']
        if m_ncdm is None:
            m_ncdm = []

        if numpy.isscalar(m_ncdm):
            # a single massive neutrino
            m_ncdm = [m_ncdm]

        if isinstance(m_ncdm, (list, numpy.ndarray)):
            m_ncdm = list(m_ncdm)
        else:
            raise TypeError("``m_ncdm`` should be a list of mass values in eV")

        for m in m_ncdm:
            if m == 0:
                raise ValueError("A zero mass is specified in the non-cold dark matter list. "
                                 "This is not needed, as we automatically set N_ur based on "
                                 "the number of entries in m_ncdm such that Neff = 3.046.")

        # number of massive neutrino species
        pars['N_ncdm'] = len(m_ncdm)

        # m_ncdm only needed if we have massive neutrinos
        if len(m_ncdm) > 0:
            pars['m_ncdm'] = m_ncdm

        # from CLASS notes:
        # one more remark: if you have respectively 1,2,3 massive neutrinos,
        # if you stick to the default value pm equal to 0.71611, designed to give m/omega of
        # 93.14 eV, and if you want to use N_ur to get N_eff equal to 3.046 in the early universe,
        # then you should pass here respectively 2.0328,1.0196,0.00641
        N_ur_table = [3.046, 2.0328, 1.0196, 0.00641]
        if args['N_ur'] is None:
            pars['N_ur'] = N_ur_table[len(m_ncdm)]

    if 'N_ur' in args:
        if args['N_ur'] is not None:
            pars['N_ur'] = args['N_ur']

    # check gauge
    if 'gauge' in args:
        if args['gauge'] not in ['synchronous', 'newtonian']:
            raise ValueError("'gauge' should be 'synchronous' or 'newtonian'")

    # set cosmological constant to zero if we got fluid w0/wa
    if 'w0_fld' in args or 'wa_fld' in args:
        if pars.get('Omega_Lambda', 0) > 0:
            raise ValueError(("non-zero Omega_Lambda (cosmological constant) specified as "
                             "well as fluid w0/wa; use Omega_fld instead"))
        pars['Omega_Lambda'] = 0.


    # maximum k value
    set_alias('P_k_max_h/Mpc', 'P_k_max')

    # maximum redshift
    set_alias('z_max_pk', 'P_z_max')

    # nonlinear
    set_alias('non linear', 'nonlinear')
    # sorry we use a boolean but
    # class uses existence of string.
    if pars.pop('non linear', False):
        pars['non linear'] = 'halofit'

    # remove None's for remaining parameters -- None means using a default from CLASS
    # NOTE: do this last since m_ncdm=None means no massive_neutrinos
    for key in list(pars.keys()):
        if pars[key] is None: pars.pop(key)

    return pars


def merge_args(args, moreargs):
    """
    merge moreargs into args.

    Those defined in moreargs takes priority than those
    defined in args.

    see :func:`compile_args`
    """
    args = args.copy()

    for name in moreargs.keys():
        # pop those conflicting with me from the old pars
        for eq in find_eqcls(name):
            if eq in args: args.pop(eq)

    args.update(moreargs)
    return args

def check_deprecated_init(args, kwargs):
    """
    Check if ``kwargs`` uses the (now deprecated) signature of ``Cosmology``
    prior to version 0.2.6.

    If using the deprecated syntax, this returns the necessary arguments for
    the new signature, and ``None`` otherwise.
    """
    from astropy import cosmology, units
    defaults = {'H0':67.6, 'Om0':0.31, 'Ob0':0.0486, 'Ode0':0.69, 'w0':-1.,
                'Tcmb0':2.7255, 'Neff':3.04, 'm_nu':0., 'flat':False}

    # the deprecated kwargs
    deprecated_args = [k for k in kwargs if k in defaults]

    # all clear; nothing to do
    if not len(deprecated_args):
        return

    # if we got deprecated kwargs, make sure we didn't get any valid kwargs!!
    if not all(a in defaults for a in kwargs) or len(kwargs) and len(args):
        msg = "mixing deprecated and valid arguments for the Cosmology class; "
        msg += 'the following args are deprecated: %s' % str(deprecated_args)
        raise ValueError(msg)

    # update old defaults with input params
    defaults.update(kwargs)

    if defaults['m_nu'] is not None:
        defaults['m_nu'] = units.Quantity(defaults['m_nu'], 'eV')

    # determine the astropy class
    if defaults['w0'] == -1.0: # cosmological constant
        cls = 'LambdaCDM'
        defaults.pop('w0')
    else:
        cls = 'wCDM'

    # use special flat case if Ok0 = 0
    if defaults.pop('flat'):
        cls = 'Flat' + cls
        defaults.pop('Ode0')

    # initialize the astropy engine and convert to dict for Cosmology()
    astropy_cosmo = getattr(cosmology, cls)(**defaults)
    return astropy_to_dict(astropy_cosmo)


def check_args(args):
    cf = {}
    for name in args.keys():
        cf[name] = []
        for eq in find_eqcls(name):
            if eq == name: continue
            if eq in args: cf[name].append(eq)

    for name in cf.keys():
        if len(cf[name]) > 0:
            raise ValueError("Conflicted parameters are given: %s" % str(cf))

# dict that defines input parameters that conflict with each other
CONFLICTS = [('h', 'H0', '100*theta_s'),
             ('T_cmb', 'Omega_g', 'omega_g', 'Omega0_g'),
             ('Omega_b', 'omega_b', 'Omega0_b'),
             ('Omega_fld', 'Omega0_fld'),
             ('Omega_Lambda', 'Omega0_Lambda'),
             ('N_ur', 'Omega_ur', 'omega_ur', 'Omega0_ur'),
             ('Omega_cdm', 'omega_cdm', 'Omega0_cdm'),
             ('m_ncdm', 'Omega_ncdm', 'omega_ncdm', 'Omega0_ncdm'),
             ('P_k_max', 'P_k_max_h/Mpc', 'P_k_max_1/Mpc'),
             ('P_z_max', 'z_max_pk'),
             ('nonlinear', 'non linear'),
             ('A_s', 'ln10^{10}A_s'),
            ]

def find_eqcls(key):
    for cls in CONFLICTS:
        if key in cls:
            return cls
    else:
        return ()
